from ._alm import _AsyncAlmClient
from ._audit import _AsyncAuditClient
from ._bulk import AsyncClient
from ._cloud_works import _AsyncCloudWorksClient
from ._cw_flow import _AsyncFlowClient
from ._scim import _AsyncScimClient
from ._transactional import _AsyncTransactionalClient

__all__ = [
    "AsyncClient",
    "_AsyncAlmClient",
    "_AsyncAuditClient",
    "_AsyncCloudWorksClient",
    "_AsyncFlowClient",
    "_AsyncTransactionalClient",
    "_AsyncScimClient",
]



from itertools import chain
from typing import Any, Literal, Type, TypeVar

from pydantic.alias_generators import to_camel

from anaplan_sdk._services import logger
from anaplan_sdk.exceptions import AnaplanException, InvalidIdentifierException
from anaplan_sdk.models import (
    AnaplanModel,
    InsertionResult,
    ModelCalendar,
    MonthsQuartersYearsCalendar,
    WeeksGeneralCalendar,
    WeeksGroupingCalendar,
    WeeksPeriodsCalendar,
)
from anaplan_sdk.models.cloud_works import (
    AmazonS3ConnectionInput,
    AzureBlobConnectionInput,
    ConnectionBody,
    GoogleBigQueryConnectionInput,
    IntegrationInput,
    IntegrationProcessInput,
    ScheduleInput,
)

T = TypeVar("T", bound=AnaplanModel)


def models_url(only_in_workspace: bool | str, workspace_id: str | None) -> str:
    if isinstance(only_in_workspace, bool) and only_in_workspace:
        if not workspace_id:
            raise ValueError(
                "Cannot list models in the current workspace, since no workspace Id was "
                "provided when instantiating the client. Either provide a workspace Id when "
                "instantiating the client, or pass a specific workspace Id to this method."
            )
        return f"https://api.anaplan.com/2/0/workspaces/{workspace_id}/models"
    if isinstance(only_in_workspace, str):
        return f"https://api.anaplan.com/2/0/workspaces/{only_in_workspace}/models"
    return "https://api.anaplan.com/2/0/models"


def sort_params(sort_by: str | None, descending: bool) -> dict[str, str | bool]:
    """
    Construct search parameters for sorting. This also converts snake_case to camelCase.
    :param sort_by: The field to sort by, optionally in snake_case.
    :param descending: Whether to sort in descending order.
    :return: A dictionary of search parameters in Anaplan's expected format.
    """
    if not sort_by:
        return {}
    return {"sort": f"{'-' if descending else '+'}{to_camel(sort_by)}"}


def construct_payload(model: Type[T], body: T | dict[str, Any]) -> dict[str, Any]:
    """
    Construct a payload for the given model and body.
    :param model: The model class to use for validation.
    :param body: The body to validate and optionally convert to a dictionary.
    :return: A dictionary representation of the validated body.
    """
    if isinstance(body, dict):
        body = model.model_validate(body)
    return body.model_dump(exclude_none=True, by_alias=True)


def connection_body_payload(body: ConnectionBody | dict[str, Any]) -> dict[str, Any]:
    """
    Construct a payload for the given integration body.
    :param body: The body to validate and optionally convert to a dictionary.
    :return: A dictionary representation of the validated body.
    """
    if isinstance(body, dict):
        if "sasToken" in body:
            body = AzureBlobConnectionInput.model_validate(body)
        elif "secretAccessKey" in body:
            body = AmazonS3ConnectionInput.model_validate(body)
        else:
            body = GoogleBigQueryConnectionInput.model_validate(body)
    return body.model_dump(exclude_none=True, by_alias=True)


def integration_payload(
    body: IntegrationInput | IntegrationProcessInput | dict[str, Any],
) -> dict[str, Any]:
    """
    Construct a payload for the given integration body.
    :param body: The body to validate and optionally convert to a dictionary.
    :return: A dictionary representation of the validated body.
    """
    if isinstance(body, dict):
        body = (
            IntegrationInput.model_validate(body)
            if "jobs" in body
            else IntegrationProcessInput.model_validate(body)
        )
    return body.model_dump(exclude_none=True, by_alias=True)


def schedule_payload(
    integration_id: str, schedule: ScheduleInput | dict[str, Any]
) -> dict[str, Any]:
    """
    Construct a payload for the given integration ID and schedule.
    :param integration_id: The ID of the integration.
    :param schedule: The schedule to validate and optionally convert to a dictionary.
    :return: A dictionary representation of the validated schedule.
    """
    if isinstance(schedule, dict):
        schedule = ScheduleInput.model_validate(schedule)
    return {
        "integrationId": integration_id,
        "schedule": schedule.model_dump(exclude_none=True, by_alias=True),
    }


def action_url(action_id: int) -> Literal["imports", "exports", "actions", "processes"]:
    """
    Determine the type of action based on its identifier.
    :param action_id: The identifier of the action.
    :return: The type of action.
    """
    if 12000000000 <= action_id < 113000000000:
        return "imports"
    if 116000000000 <= action_id < 117000000000:
        return "exports"
    if 117000000000 <= action_id < 118000000000:
        return "actions"
    if 118000000000 <= action_id < 119000000000:
        return "processes"
    raise InvalidIdentifierException(f"Action '{action_id}' is not a valid identifier.")


def parse_calendar_response(data: dict) -> ModelCalendar:
    """
    Parse calendar response and return appropriate calendar model.
    :param data: The calendar data from the API response.
    :return: The calendar settings of the model based on calendar type.
    """
    calendar_data = data["modelCalendar"]
    cal_type = calendar_data["calendarType"]
    if cal_type == "Calendar Months/Quarters/Years":
        return MonthsQuartersYearsCalendar.model_validate(calendar_data)
    if cal_type == "Weeks: 4-4-5, 4-5-4 or 5-4-4":
        return WeeksGroupingCalendar.model_validate(calendar_data)
    if cal_type == "Weeks: General":
        return WeeksGeneralCalendar.model_validate(calendar_data)
    if cal_type == "Weeks: 13 4-week Periods":
        return WeeksPeriodsCalendar.model_validate(calendar_data)
    raise AnaplanException(
        "Unknown calendar type encountered. Please report this issue: "
        "https://github.com/VinzenzKlass/anaplan-sdk/issues/new"
    )


def parse_insertion_response(data: list[dict]) -> InsertionResult:
    failures, added, ignored, total = [], 0, 0, 0
    for res in data:
        failures.append(res.get("failures", []))
        added += res.get("added", 0)
        total += res.get("total", 0)
        ignored += res.get("ignored", 0)
    return InsertionResult(
        added=added, ignored=ignored, total=total, failures=list(chain.from_iterable(failures))
    )


def validate_dimension_id(dimension_id: int) -> int:
    if not (
        dimension_id == 101999999999
        or 101_000_000_000 <= dimension_id < 102_000_000_000
        or 109_000_000_000 <= dimension_id < 110_000_000_000
        or 114_000_000_000 <= dimension_id < 115_000_000_000
    ):
        raise InvalidIdentifierException(
            "Invalid dimension_id. Must be a List (101xxxxxxxxx), List Subset (109xxxxxxxxx), "
            "Line Item Subset (114xxxxxxxxx), or Users (101999999999)."
        )
    msg = (
        "Using `get_dimension_items` for {} is discouraged. "
        "Prefer `{}` for better performance and more details on the members."
    )
    if dimension_id == 101999999999:
        logger.warning(msg.format("Users", "get_users"))
    if 101000000000 <= dimension_id < 102000000000:
        logger.warning(msg.format("Lists", "get_list_items"))
    return dimension_id



from typing import Any, Literal

from pydantic import Field
from typing_extensions import Self

from anaplan_sdk.models import AnaplanModel

AnaplanFilterFields = Literal[
    "id", "externalId", "userName", "name.familyName", "name.givenName", "active"
]


class _FilterExpression:
    def __init__(self, _field: AnaplanFilterFields) -> None:
        self._field: AnaplanFilterFields = _field
        self._exprs: list[str] = []
        self._operators = []

    def __str__(self) -> str:
        if not self._exprs:
            return "active eq true" if self._field == "active" else f"{self._field} pr"
        parts = []
        for i, expr in enumerate(self._exprs):
            if i > 0:
                parts.append(self._operators[i - 1])
            parts.append(expr)
        return " ".join(parts)

    def __invert__(self) -> Self:
        self._exprs.append(
            "active eq false" if self._field == "active" else f"{self._field} eq null"
        )
        return self

    def __and__(self, other: Self) -> Self:
        return self._combine(other, "and")

    def __or__(self, other: Self) -> Self:
        return self._combine(other, "or")

    def _combine(self, other: Self, operator: Literal["and", "or"]) -> Self:
        if self._requires_grouping(operator):
            self._exprs = [f"({str(self)})"]
            self._operators = []
        self._operators.append(operator)
        if not self._exprs:
            self._exprs.append(str(self))
        self._exprs.append(f"({str(other)})" if other._requires_grouping(operator) else str(other))
        return self

    def _requires_grouping(self, operator: Literal["and", "or"]) -> bool:
        return len(self._operators) > 0 and ("or" in self._operators or operator == "or")

    def __eq__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} eq "{other}"')
        return self

    def __ne__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} ne "{other}"')
        return self

    def __gt__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} gt "{other}"')
        return self

    def __ge__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} ge "{other}"')
        return self

    def __lt__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} lt "{other}"')
        return self

    def __le__(self, other: Any) -> Self:
        self._exprs.append(f'{self._field} le "{other}"')
        return self


field = _FilterExpression


class NameInput(AnaplanModel):
    family_name: str = Field(
        description="The family name of the User, or last name in most Western languages"
    )
    given_name: str = Field(
        description="The given name of the User, or first name in most Western languages"
    )


class Name(NameInput):
    formatted: str = Field(
        description=(
            "The formatted full name, including given name and family name. Anaplan does as of now "
            "not have other standard SCIM fields such as middle name or honorific pre- or suffixes."
        )
    )


class Email(AnaplanModel):
    value: str = Field(description="Email address of the User")
    type: Literal["work", "home", "other"] = Field(
        default=None, description="A label indicating the emails's function, e.g., 'work' or 'home'"
    )
    primary: bool | None = Field(
        default=None, description="Indicates if this is the primary or 'preferred' email"
    )


class EntitlementInput(AnaplanModel):
    value: str = Field(description="The value of an entitlement.")
    type: Literal["WORKSPACE", "WORKSPACE_IDS", "WORKSPACE_NAMES"] = Field(
        description="A label indicating the attribute's function."
    )


class Entitlement(EntitlementInput):
    display: str | None = Field(
        default=None, description="A human-readable name, primarily used for display purposes."
    )
    primary: bool | None = Field(
        default=None,
        description="Indicating the 'primary' or preferred attribute value for this attribute.",
    )


class _BaseUser(AnaplanModel):
    schemas: list[str] = Field(default=["urn:ietf:params:scim:schemas:core:2.0:User"])
    user_name: str = Field(description="Unique name for the User.")


class User(_BaseUser):
    id: str = Field(description="The unique identifier for the User.")
    name: Name = Field(description="The user's real name.")
    active: bool = Field(description="Indicating the User's active status.")
    emails: list[Email] = Field(default=[], description="Email addresses for the user.")
    display_name: str = Field(description="Display Name for the User.")
    entitlements: list[Entitlement] = Field(
        default=[], description="A list of entitlements (Workspaces) the User has."
    )


class ReplaceUserInput(_BaseUser):
    id: str = Field(description="The unique identifier for the User.")
    name: NameInput = Field(description="The user's real name.")
    active: bool | None = Field(default=None, description="Indicating the User's active status.")
    display_name: str | None = Field(default=None, description="Display Name for the User.")
    entitlements: list[EntitlementInput] | None = Field(
        default=None, description="A list of entitlements (Workspaces) the User has."
    )


class UserInput(_BaseUser):
    external_id: str = Field(
        description="Your unique id for this user (as stored in your company systems)."
    )
    name: NameInput = Field(description="The user's real name.")


class Meta(AnaplanModel):
    resource_type: str = Field(description="The type of the resource.")
    location: str = Field(description="The URI of the resource.")


class MetaWithDates(Meta):
    created: str = Field(description="The timestamp when the resource was created.")
    last_modified: str = Field(description="The timestamp when the resource was last modified.")


class Supported(AnaplanModel):
    supported: bool = Field(description="Indicates whether the Feature is supported.")


class BulkConfig(Supported):
    max_operations: int = Field(
        description="The maximum number of operations permitted in a single request."
    )
    max_payload_size: int = Field(description="The maximum payload size in bytes.")


class FilterConfig(Supported):
    max_results: int = Field(
        description="The maximum number of results returned from a filtered query."
    )


class AuthenticationScheme(AnaplanModel):
    name: str = Field(description="The name of the authentication scheme.")
    type: str = Field(description="The type of the authentication scheme.")
    description: str = Field(description="A description of the authentication scheme.")
    spec_uri: str = Field(
        description="The URI that points to the specification of the authentication scheme."
    )
    documentation_uri: str = Field(
        description="The URI that points to the documentation of the authentication scheme."
    )


class ServiceProviderConfig(AnaplanModel):
    schemas: list[str] = Field(
        default=["urn:ietf:params:scim:schemas:core:2.0:ServiceProviderConfig"],
        description="Schemas for this resource.",
    )
    meta: MetaWithDates = Field(description="Metadata about the resource.")
    documentation_uri: str = Field(description="URI of the service provider's documentation.")
    patch: Supported = Field(description="Configuration for PATCH operations.")
    bulk: BulkConfig = Field(description="Configuration for bulk operations.")
    filter: FilterConfig = Field(description="Configuration for filtering.")
    change_password: Supported = Field(description="Configuration for password changes.")
    sort: Supported = Field(description="Configuration for sorting.")
    etag: Supported = Field(description="Configuration for ETags.")
    authentication_schemes: list[AuthenticationScheme] = Field(
        description="List of supported authentication schemes."
    )


class Resource(AnaplanModel):
    schemas: list[str] = Field(
        default=["urn:ietf:params:scim:schemas:core:2.0:ResourceType"],
        description="Schemas for this resource.",
    )
    meta: Meta = Field(description="Metadata about the resource.")
    id: str = Field(description="The identifier of the resource type.")
    name: str = Field(description="The name of the resource type.")
    endpoint: str = Field(description="The endpoint where resources of this type may be accessed.")
    description: str = Field(description="A description of the resource type.")


class Attribute(AnaplanModel):
    name: str = Field(description="The name of the attribute.")
    type: str = Field(description="The data type of the attribute.")
    multi_valued: bool = Field(description="Indicates if the attribute can have multiple values.")
    description: str = Field(description="A human-readable description of the attribute.")
    required: bool = Field(description="Indicates if the attribute is required.")
    case_exact: bool = Field(
        description="Indicates if case sensitivity should be considered when comparing values."
    )
    mutability: str = Field(description="Indicates if and how the attribute can be modified.")
    returned: str = Field(
        description="Indicates when the attribute's values are returned in a response."
    )
    uniqueness: str = Field(
        description="Indicates how uniqueness is enforced on the attribute value."
    )
    sub_attributes: list["Attribute"] | None = Field(
        default=None, description="A list of sub-attributes if the attribute is complex."
    )


class Schema(AnaplanModel):
    meta: Meta = Field(description="Metadata about the schema resource.")
    id: str = Field(description="The unique identifier for the schema.")
    name: str = Field(description="The name of the schema.")
    description: str = Field(description="A description of the schema.")
    attributes: list[Attribute] = Field(description="A list of attributes that define the schema.")


class Operation(AnaplanModel):
    op: Literal["add", "remove", "replace"] = Field(description="The operation to be performed.")
    path: str = Field(
        description=(
            "A string containing a JSON-Pointer value that references a location within the target "
            "resource."
        )
    )
    value: Any | None = Field(default=None, description="The value to be used in the operation.")


class Replace(Operation):
    op: Literal["replace"] = Field(
        default="replace", description="Replace the value at path with the new given value."
    )


class Add(Operation):
    op: Literal["add"] = Field(
        default="add", description="Add the given value to the attribute at path."
    )


class Remove(Operation):
    op: Literal["remove"] = Field(default="remove", description="Remove the value at path.")



import asyncio
import logging
import time
from asyncio import gather, sleep
from concurrent.futures import ThreadPoolExecutor
from gzip import compress
from itertools import chain
from math import ceil
from typing import Any, Awaitable, Callable, Coroutine, Iterator, TypeVar

import httpx
from httpx import HTTPError, Response

from .exceptions import AnaplanException, AnaplanTimeoutException, InvalidIdentifierException
from .models import TaskSummary

SORT_WARNING = (
    "If you are sorting by a field that is potentially ambiguous (e.g., name), the order of "
    "results is not guaranteed to be internally consistent across multiple requests. This will "
    "lead to wrong results when paginating through result sets where the ambiguous order can cause "
    "records to slip between pages or be duplicated on multiple pages. The only way to ensure "
    "correct results when sorting is to make sure the entire result set fits in one page, or to "
    "sort by a field that is guaranteed to be unique (e.g., id)."
)

logger = logging.getLogger("anaplan_sdk")

_json_header = {"Content-Type": "application/json"}
_gzip_header = {"Content-Type": "application/x-gzip"}

Task = TypeVar("Task", bound=TaskSummary)


class _HttpService:
    def __init__(
        self,
        client: httpx.Client,
        *,
        retry_count: int,
        backoff: float,
        backoff_factor: float,
        page_size: int,
        poll_delay: int,
    ):
        logger.debug(
            f"Initializing HttpService with retry_count={retry_count}, "
            f"page_size={page_size}, poll_delay={poll_delay}."
        )
        self._client = client
        self._retry_count = retry_count
        self._backoff = backoff
        self._backoff_factor = backoff_factor
        self._poll_delay = poll_delay
        self._page_size = min(page_size, 5_000)

    def get(self, url: str, **kwargs) -> dict[str, Any]:
        return self.__run_with_retry(self._client.get, url, **kwargs).json()

    def get_binary(self, url: str) -> bytes:
        return self.__run_with_retry(self._client.get, url).content

    def post(self, url: str, json: dict | list) -> dict[str, Any]:
        return self.__run_with_retry(self._client.post, url, headers=_json_header, json=json).json()

    def put(self, url: str, json: dict | list) -> dict[str, Any]:
        res = self.__run_with_retry(self._client.put, url, headers=_json_header, json=json)
        return res.json() if res.num_bytes_downloaded > 0 else {}

    def patch(self, url: str, json: dict | list) -> dict[str, Any]:
        return (
            self.__run_with_retry(self._client.patch, url, headers=_json_header, json=json)
        ).json()

    def delete(self, url: str) -> dict[str, Any]:
        return (self.__run_with_retry(self._client.delete, url, headers=_json_header)).json()

    def post_empty(self, url: str, **kwargs) -> dict[str, Any]:
        res = self.__run_with_retry(self._client.post, url, **kwargs)
        return res.json() if res.num_bytes_downloaded > 0 else {}

    def put_binary_gzip(self, url: str, content: str | bytes) -> Response:
        content = compress(content.encode() if isinstance(content, str) else content)
        return self.__run_with_retry(self._client.put, url, headers=_gzip_header, content=content)

    def poll_task(self, func: Callable[..., Task], *args) -> Task:
        while (result := func(*args)).task_state != "COMPLETE":
            time.sleep(self._poll_delay)
        return result

    def get_paginated(self, url: str, result_key: str, **kwargs) -> Iterator[dict[str, Any]]:
        logger.debug(f"Starting paginated fetch from {url} with page_size={self._page_size}.")
        first_page, total_items, actual_size = self._get_first_page(url, result_key, **kwargs)
        if total_items <= self._page_size:
            logger.debug("All items fit in first page, no additional requests needed.")
            return iter(first_page)
        if kwargs and (kwargs.get("params") or {}).get("sort", None):
            logger.warning(SORT_WARNING)
        pages_needed = ceil(total_items / actual_size)
        logger.debug(f"Fetching {pages_needed - 1} additional pages with {actual_size} items each.")
        with ThreadPoolExecutor() as executor:
            pages = executor.map(
                lambda n: self._get_page(url, actual_size, n * actual_size, result_key, **kwargs),
                range(1, pages_needed),
            )
        logger.debug(f"Completed paginated fetch of {total_items} total items.")
        return chain(first_page, *pages)

    def _get_page(self, url: str, limit: int, offset: int, result_key: str, **kwargs) -> list:
        logger.debug(f"Fetching page: offset={offset}, limit={limit} from {url}.")
        kwargs["params"] = (kwargs.get("params") or {}) | {"limit": limit, "offset": offset}
        return self.get(url, **kwargs).get(result_key, [])

    def _get_first_page(self, url: str, result_key: str, **kwargs) -> tuple[list, int, int]:
        logger.debug(f"Fetching first page with limit={self._page_size} from {url}.")
        kwargs["params"] = (kwargs.get("params") or {}) | {"limit": self._page_size}
        res = self.get(url, **kwargs)
        return _extract_first_page(res, result_key, self._page_size)

    def __run_with_retry(self, func: Callable[..., Response], *args, **kwargs) -> Response:
        for i in range(max(self._retry_count, 1)):
            try:
                response = func(*args, **kwargs)
                if response.status_code == 429:
                    if i >= self._retry_count - 1:
                        raise AnaplanException("Rate limit exceeded.")
                    backoff_time = self._backoff * (self._backoff_factor if i > 0 else 1)
                    logger.warning(f"Rate limited. Retrying in {backoff_time} seconds.")
                    time.sleep(backoff_time)
                    continue
                response.raise_for_status()
                return response
            except HTTPError as error:
                if i >= self._retry_count - 1:
                    _raise_error(error)
                url = args[0] or kwargs.get("url")
                logger.info(f"Retrying for: {url}")

        raise AnaplanException("Exhausted all retries without a successful response or Error.")


class _AsyncHttpService:
    def __init__(
        self,
        client: httpx.AsyncClient,
        *,
        retry_count: int,
        backoff: float,
        backoff_factor: float,
        page_size: int,
        poll_delay: int,
    ):
        logger.debug(
            f"Initializing AsyncHttpService with retry_count={retry_count}, "
            f"page_size={page_size}, poll_delay={poll_delay}."
        )
        self._client = client
        self._retry_count = retry_count
        self._backoff = backoff
        self._backoff_factor = backoff_factor
        self._poll_delay = poll_delay
        self._page_size = min(page_size, 5_000)

    async def get(self, url: str, **kwargs) -> dict[str, Any]:
        return (await self._run_with_retry(self._client.get, url, **kwargs)).json()

    async def get_binary(self, url: str) -> bytes:
        return (await self._run_with_retry(self._client.get, url)).content

    async def post(self, url: str, json: dict | list) -> dict[str, Any]:
        return (
            await self._run_with_retry(self._client.post, url, headers=_json_header, json=json)
        ).json()

    async def put(self, url: str, json: dict | list) -> dict[str, Any]:
        res = await self._run_with_retry(self._client.put, url, headers=_json_header, json=json)
        return res.json() if res.num_bytes_downloaded > 0 else {}

    async def patch(self, url: str, json: dict | list) -> dict[str, Any]:
        return (
            await self._run_with_retry(self._client.patch, url, headers=_json_header, json=json)
        ).json()

    async def delete(self, url: str) -> dict[str, Any]:
        return (await self._run_with_retry(self._client.delete, url, headers=_json_header)).json()

    async def post_empty(self, url: str, **kwargs) -> dict[str, Any]:
        res = await self._run_with_retry(self._client.post, url, **kwargs)
        return res.json() if res.num_bytes_downloaded > 0 else {}

    async def put_binary_gzip(self, url: str, content: str | bytes) -> Response:
        content = compress(content.encode() if isinstance(content, str) else content)
        return await self._run_with_retry(
            self._client.put, url, headers=_gzip_header, content=content
        )

    async def poll_task(self, func: Callable[..., Awaitable[Task]], *args) -> Task:
        while (result := await func(*args)).task_state != "COMPLETE":
            await sleep(self._poll_delay)
        return result

    async def get_paginated(self, url: str, result_key: str, **kwargs) -> Iterator[dict[str, Any]]:
        logger.debug(f"Starting paginated fetch from {url} with page_size={self._page_size}.")
        first_page, total_items, actual_size = await self._get_first_page(url, result_key, **kwargs)
        if total_items <= self._page_size:
            logger.debug("All items fit in first page, no additional requests needed.")
            return iter(first_page)
        if kwargs and (kwargs.get("params") or {}).get("sort", None):
            logger.warning(SORT_WARNING)
        pages = await gather(
            *(
                self._get_page(url, actual_size, n * actual_size, result_key, **kwargs)
                for n in range(1, ceil(total_items / actual_size))
            )
        )
        logger.debug(f"Completed paginated fetch of {total_items} total items.")
        return chain(first_page, *pages)

    async def _get_page(self, url: str, limit: int, offset: int, result_key: str, **kwargs) -> list:
        logger.debug(f"Fetching page: offset={offset}, limit={limit} from {url}.")
        kwargs["params"] = (kwargs.get("params") or {}) | {"limit": limit, "offset": offset}
        return (await self.get(url, **kwargs)).get(result_key, [])

    async def _get_first_page(self, url: str, result_key: str, **kwargs) -> tuple[list, int, int]:
        logger.debug(f"Fetching first page with limit={self._page_size} from {url}.")
        kwargs["params"] = (kwargs.get("params") or {}) | {"limit": self._page_size}
        res = await self.get(url, **kwargs)
        return _extract_first_page(res, result_key, self._page_size)

    async def _run_with_retry(
        self, func: Callable[..., Coroutine[Any, Any, Response]], *args, **kwargs
    ) -> Response:
        for i in range(max(self._retry_count, 1)):
            try:
                response = await func(*args, **kwargs)
                if response.status_code == 429:
                    if i >= self._retry_count - 1:
                        raise AnaplanException("Rate limit exceeded.")
                    backoff_time = self._backoff * (self._backoff_factor if i > 0 else 1)
                    logger.warning(f"Rate limited. Retrying in {backoff_time} seconds.")
                    await asyncio.sleep(backoff_time)
                    continue
                response.raise_for_status()
                return response
            except HTTPError as error:
                if i >= self._retry_count - 1:
                    _raise_error(error)
                url = args[0] or kwargs.get("url")
                logger.info(f"Retrying for: {url}")

        raise AnaplanException("Exhausted all retries without a successful response or Error.")


def _extract_first_page(
    res: dict[str, Any], result_key: str, page_size: int
) -> tuple[list[dict[str, Any]], int, int]:
    total_items, first_page = res["meta"]["paging"]["totalSize"], res.get(result_key, [])
    actual_page_size = res["meta"]["paging"]["currentPageSize"]
    if actual_page_size < page_size and not actual_page_size == total_items:
        logger.warning(
            f"Page size {page_size} was silently truncated to {actual_page_size}."
            f"Using the server-side enforced page size {actual_page_size} for further requests."
        )
    logger.debug(f"Found {total_items} total items, retrieved {len(first_page)} in first page.")
    return first_page, total_items, actual_page_size


def _raise_error(error: HTTPError) -> None:
    if isinstance(error, httpx.TimeoutException):
        raise AnaplanTimeoutException from error
    if isinstance(error, httpx.HTTPStatusError):
        if error.response.status_code == 404:
            raise InvalidIdentifierException from error
        logger.error(f"Anaplan Error: [{error.response.status_code}]: {error.response.text}")
        raise AnaplanException(error.response.text) from error

    logger.error(f"Error: {error}")
    raise AnaplanException from error



from ._alm import (
    ModelRevision,
    ReportTask,
    ReportTaskResult,
    Revision,
    SummaryReport,
    SyncTask,
    SyncTaskResult,
)
from ._base import AnaplanModel
from ._bulk import (
    Action,
    Export,
    ExportTypes,
    File,
    Import,
    ImportTypes,
    List,
    ListMetadata,
    Model,
    ModelDeletionResult,
    Process,
    TaskResult,
    TaskResultDetail,
    TaskStatus,
    TaskSummary,
    Workspace,
)
from ._transactional import (
    CurrentPeriod,
    Dimension,
    DimensionWithCode,
    Failure,
    FiscalYear,
    InsertionResult,
    LineItem,
    ListDeletionResult,
    ListItem,
    ModelCalendar,
    ModelStatus,
    Module,
    MonthsQuartersYearsCalendar,
    User,
    View,
    ViewInfo,
    WeeksGeneralCalendar,
    WeeksGroupingCalendar,
    WeeksPeriodsCalendar,
)

__all__ = [
    "AnaplanModel",
    "ExportTypes",
    "ImportTypes",
    "Workspace",
    "Model",
    "ModelStatus",
    "ModelRevision",
    "File",
    "List",
    "ListItem",
    "ListMetadata",
    "Action",
    "Import",
    "Export",
    "Process",
    "Module",
    "LineItem",
    "TaskSummary",
    "TaskResult",
    "TaskResultDetail",
    "TaskStatus",
    "TaskSummary",
    "SyncTaskResult",
    "ReportTask",
    "ReportTaskResult",
    "SummaryReport",
    "SyncTask",
    "User",
    "Failure",
    "InsertionResult",
    "Revision",
    "CurrentPeriod",
    "FiscalYear",
    "MonthsQuartersYearsCalendar",
    "WeeksGeneralCalendar",
    "WeeksGroupingCalendar",
    "WeeksPeriodsCalendar",
    "Dimension",
    "View",
    "ViewInfo",
    "ModelCalendar",
    "ModelDeletionResult",
    "DimensionWithCode",
    "ListDeletionResult",
]



from typing import Literal, TypeAlias

from pydantic import Field, field_validator

from ._base import AnaplanModel

ExportTypes: TypeAlias = Literal[
    "TABULAR_MULTI_COLUMN",
    "TABULAR_SINGLE_COLUMN",
    "GRID_CURRENT_PAGE",
    "AUDIT_LOG",
    "TABULAR_ALL_LINE_ITEMS",
    "TABULAR_CURRENT_LINE_ITEM",
]

ImportTypes: TypeAlias = Literal[
    "MODULE_DATA", "HIERARCHY_DATA", "LINE_ITEM_DEFINITION", "USERS", "VERSIONS"
]


class Workspace(AnaplanModel):
    id: str = Field(description="The unique identifier of this workspace.")
    name: str = Field(description="The name of this workspace that is also displayed to the users.")
    active: bool = Field(description="Whether this workspace is active or not.")
    size_allowance: int = Field(description="The maximum allowed size of this workspace in bytes.")
    current_size: int = Field(description="The current size of this workspace in bytes.")


class Model(AnaplanModel):
    id: str = Field(description="The unique identifier of this model.")
    name: str
    active_state: Literal[
        "ARCHIVED", "UNLOCKED", "ACTIVE", "PRODUCTION", "MAINTENANCE", "PRODUCTION_MAINTENANCE"
    ] = Field(description="The current state of this model.")
    last_saved_serial_number: int = Field(
        description="The serial number of the last save of this model."
    )
    last_modified_by_user_guid: str = Field(
        description="The unique identifier of the user who last modified this model."
    )
    memory_usage: int = Field(0, description="The memory usage of this model in bytes.")
    workspace_id: str = Field(
        validation_alias="currentWorkspaceId",
        description="The unique identifier of the workspace that this model is currently in.",
    )
    workspace_name: str = Field(
        validation_alias="currentWorkspaceName",
        description="The name of the workspace that this model is currently in.",
    )
    url: str = Field(validation_alias="modelUrl", description="The current URL of this model.")
    category_values: list = Field(description="The category values of this model.")
    iso_creation_date: str = Field(description="The creation date of this model in ISO format.")
    last_modified: str = Field(description="The last modified date of this model.")


class File(AnaplanModel):
    id: int = Field(description="The unique identifier of this file.")
    name: str = Field(description="The name of this file.")
    chunk_count: int = Field(description="The number of chunks this file is split into.")
    delimiter: str | None = Field(None, description="The delimiter used in this file.")
    encoding: str | None = Field(None, description="The encoding of this file.")
    first_data_row: int = Field(description="The row number of the first data row in this file.")
    format: str | None = Field(None, description="The format of this file.")
    header_row: int = Field(description="The row number of the header row in this file.")
    separator: str | None = Field(None, description="The separator used in this file.")


class List(AnaplanModel):
    id: int = Field(description="The unique identifier of this list.")
    name: str = Field(description="The name of this list.")


class ListMetadata(AnaplanModel):
    id: int = Field(description="The unique identifier of this list.")
    name: str = Field(description="The name of this list.")
    has_selective_access: bool = Field(description="Whether this list has selective access or not.")
    properties: list = Field([], description="The properties of this list.")
    production_data: bool = Field(description="Whether this list is production data or not.")
    managed_by: str = Field(description="The user who manages this list.")
    numbered_list: bool = Field(description="Whether this list is a numbered list or not.")
    use_top_level_as_page_default: bool = Field(
        description="Whether the top level is used as the page default or not."
    )
    item_count: int = Field(description="The number of items in this list.")
    next_item_index: int | None = Field(
        None, description="The index of the next item in this list."
    )
    workflow_enabled: bool = Field(
        description="Whether the workflow is enabled for this list or not."
    )
    permitted_items: int = Field(description="The number of permitted items in this list.")
    used_in_applies_to: str | None = Field(None, description="The applies to value of this list.")


class Action(AnaplanModel):
    id: int = Field(description="The unique identifier of this action.")
    name: str = Field(
        description="The name of this Action. This is the same as the one displayed in the Web UI."
    )
    type: str | None = Field(
        None, validation_alias="actionType", description="The type of this action."
    )


class Process(AnaplanModel):
    id: int = Field(description="The unique identifier of this process.")
    name: str = Field(description="The name of this process.")


class Import(AnaplanModel):
    id: int = Field(description="The unique identifier of this import.")
    name: str = Field(description="The name of this import.")
    type: ImportTypes = Field(validation_alias="importType", description="The type of this import.")
    file_id: int | None = Field(
        None,
        validation_alias="importDataSourceId",
        description=(
            "The unique identifier of the data source of this import. If it is absent, it means "
            "that the import does not read from any file."
        ),
    )

    @field_validator("file_id", mode="before")
    @classmethod
    def _empty_source_is_none(cls, inp: str):
        return inp if inp else None


class Export(AnaplanModel):
    id: int = Field(description="The unique identifier of this export.")
    name: str = Field(description="The name of this export.")
    type: ExportTypes = Field(validation_alias="exportType", description="The type of this export.")
    format: str = Field(validation_alias="exportFormat", description="The format of this export.")
    encoding: str | None = Field(None, description="The encoding of this export.")
    layout: ExportTypes = Field(
        description="The layout of this export, representing the Anaplan Export Structure."
    )


class TaskSummary(AnaplanModel):
    id: str = Field(validation_alias="taskId", description="The unique identifier of this task.")
    task_state: Literal["NOT_STARTED", "IN_PROGRESS", "COMPLETE"] = Field(
        description="The state of this task."
    )
    creation_time: int = Field(description="Unix timestamp of when this task was created.")


class TaskResultDetail(AnaplanModel):
    local_message_text: str | None = Field(None, description="Error message text.")
    occurrences: int = Field(0, description="The number of occurrences of this error.")
    type: str = Field(description="The type of this error.")
    values: list[str | None] = Field([], description="Further error information if available.")


class TaskResult(AnaplanModel):
    details: list[TaskResultDetail] = Field(
        [], description="The details of this task result if available."
    )
    successful: bool = Field(description="Whether this task completed successfully or not.")
    failure_dump_available: bool = Field(
        description="Whether this task completed successfully or not."
    )
    nested_results: list["TaskResult"] = Field(
        [], description="The nested results of this task, if available."
    )


class TaskStatus(TaskSummary):
    progress: float = Field(description="The progress of this task as a float between 0 and 1.")
    current_step: str | None = Field(None, description="The current step of this task.")
    result: TaskResult | None = Field(None)


class DeletionFailure(AnaplanModel):
    model_id: str = Field(description="The unique identifier of the model that failed to delete.")
    message: str = Field(description="The error message explaining why the deletion failed.")


class ModelDeletionResult(AnaplanModel):
    models_deleted: int = Field(description="The number of models that were successfully deleted.")
    failures: list[DeletionFailure] = Field(
        [],
        validation_alias="bulkDeleteModelsFailures",
        description="List of models that failed to delete with their error messages.",
    )



import logging
from typing import Literal, overload

from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import sort_params
from anaplan_sdk.exceptions import AnaplanActionError
from anaplan_sdk.models import (
    ModelRevision,
    ReportTask,
    Revision,
    SummaryReport,
    SyncTask,
    TaskSummary,
)

logger = logging.getLogger("anaplan_sdk")


class _AlmClient:
    def __init__(self, http: _HttpService, model_id: str) -> None:
        self._http = http
        self._model_id = model_id
        self._url = f"https://api.anaplan.com/2/0/models/{model_id}"

    def change_model_status(self, status: Literal["online", "offline"]) -> None:
        """
        Use this call to change the status of a model.
        :param status: The status of the model. Can be either "online" or "offline".
        """
        logger.info(f"Changed model status to '{status}' for model {self._model_id}.")
        self._http.put(f"{self._url}/onlineStatus", json={"status": status})

    def get_revisions(
        self,
        sort_by: Literal["id", "name", "applied_on", "created_on"] | None = None,
        descending: bool = False,
    ) -> list[Revision]:
        """
        Use this call to return a list of revisions for a specific model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: A list of revisions for a specific model.
        """
        res = self._http.get_paginated(
            f"{self._url}/alm/revisions", "revisions", params=sort_params(sort_by, descending)
        )
        return [Revision.model_validate(e) for e in res]

    def get_latest_revision(self) -> Revision | None:
        """
        Use this call to return the latest revision for a specific model. The response is in the
        same format as in Getting a list of syncable revisions between two models.

        If a revision exists, the return list should contain one element only which is the
        latest revision.
        :return: The latest revision for a specific model, or None if no revisions exist.
        """
        res = (self._http.get(f"{self._url}/alm/latestRevision")).get("revisions")
        return Revision.model_validate(res[0]) if res else None

    def get_syncable_revisions(self, source_model_id: str) -> list[Revision]:
        """
        Use this call to return the list of revisions from your source model that can be
        synchronized to your target model.

        The returned list displays in descending order, by creation date and time. This is
        consistent with how revisions are displayed in the user interface (UI).
        :param source_model_id: The ID of the source model.
        :return: A list of revisions that can be synchronized to the target model.
        """
        res = self._http.get(f"{self._url}/alm/syncableRevisions?sourceModelId={source_model_id}")
        return [Revision.model_validate(e) for e in res.get("revisions", [])]

    def create_revision(self, name: str, description: str) -> Revision:
        """
        Create a new revision for the model.
        :param name: The name (title) of the revision.
        :param description: The description of the revision.
        :return: The created Revision Info.
        """
        res = self._http.post(
            f"{self._url}/alm/revisions", json={"name": name, "description": description}
        )
        rev = Revision.model_validate(res["revision"])
        logger.info(f"Created revision '{name} ({rev.id})'for model {self._model_id}.")
        return rev

    def get_sync_tasks(self) -> list[TaskSummary]:
        """
        List the sync tasks for a target mode. The returned the tasks are either in progress, or
        they completed within the last 48 hours.
        :return: A list of sync tasks in descending order of creation time.
        """
        res = self._http.get(f"{self._url}/alm/syncTasks")
        return [TaskSummary.model_validate(e) for e in res.get("tasks", [])]

    def get_sync_task(self, task_id: str) -> SyncTask:
        """
        Get the information for a specific sync task.
        :param task_id: The ID of the sync task.
        :return: The sync task information.
        """
        res = self._http.get(f"{self._url}/alm/syncTasks/{task_id}")
        return SyncTask.model_validate(res["task"])

    def sync_models(
        self,
        source_revision_id: str,
        source_model_id: str,
        target_revision_id: str,
        wait_for_completion: bool = True,
    ) -> SyncTask:
        """
        Create a synchronization task between two revisions. This will synchronize the
        source revision of the source model to the target revision of the target model. This will
        fail if the source revision is incompatible with the target revision.
        :param source_revision_id: The ID of the source revision.
        :param source_model_id: The ID of the source model.
        :param target_revision_id: The ID of the target revision.
        :param wait_for_completion: If True, the method will poll the task status and not return
               until the task is complete. If False, it will spawn the task and return immediately.
        :return: The created sync task.
        """
        payload = {
            "sourceRevisionId": source_revision_id,
            "sourceModelId": source_model_id,
            "targetRevisionId": target_revision_id,
        }
        res = self._http.post(f"{self._url}/alm/syncTasks", json=payload)
        task = self.get_sync_task(res["task"]["taskId"])
        logger.info(
            f"Started sync task '{task.id}' from Model '{source_model_id}' "
            f"(Revision '{source_revision_id}') to Model '{self._model_id}'."
        )
        if not wait_for_completion:
            return task
        task = self._http.poll_task(self.get_sync_task, task.id)
        if not task.result.successful:
            msg = f"Sync task {task.id} completed with errors: {task.result.error}."
            logger.error(msg)
            raise AnaplanActionError(msg)
        logger.info(f"Sync task {task.id} completed successfully.")
        return task

    def get_models_for_revision(self, revision_id: str) -> list[ModelRevision]:
        """
        Use this call when you need a list of the models that had a specific revision applied
        to them.
        :param revision_id: The ID of the revision.
        :return: A list of models that had a specific revision applied to them.
        """
        res = self._http.get(f"{self._url}/alm/revisions/{revision_id}/appliedToModels")
        return [ModelRevision.model_validate(e) for e in res.get("appliedToModels", [])]

    def create_comparison_report(
        self,
        source_revision_id: str,
        source_model_id: str,
        target_revision_id: str,
        wait_for_completion: bool = True,
    ) -> ReportTask:
        """
        Generate a full comparison report between two revisions. This will list all the changes made
        to the source revision compared to the target revision.
        :param source_revision_id: The ID of the source revision.
        :param source_model_id: The ID of the source model.
        :param target_revision_id: The ID of the target revision.
        :param wait_for_completion: If True, the method will poll the task status and not return
               until the task is complete. If False, it will spawn the task and return immediately.
        :return: The created report task summary.
        """
        payload = {
            "sourceRevisionId": source_revision_id,
            "sourceModelId": source_model_id,
            "targetRevisionId": target_revision_id,
        }
        res = self._http.post(f"{self._url}/alm/comparisonReportTasks", json=payload)
        task = self.get_comparison_report_task(res["task"]["taskId"])
        logger.info(
            f"Started Comparison Report task '{task.id}' between Model '{source_model_id}' "
            f"(Revision '{source_revision_id}') and Model '{self._model_id}'."
        )
        if not wait_for_completion:
            return task
        task = self._http.poll_task(self.get_comparison_report_task, task.id)
        if not task.result.successful:
            msg = f"Comparison Report task {task.id} completed with errors: {task.result.error}."
            logger.error(msg)
            raise AnaplanActionError(msg)
        logger.info(f"Comparison Report task {task.id} completed successfully.")
        return task

    def get_comparison_report_task(self, task_id: str) -> ReportTask:
        """
        Get the task information for a comparison report task.
        :param task_id: The ID of the comparison report task.
        :return: The report task information.
        """
        res = self._http.get(f"{self._url}/alm/comparisonReportTasks/{task_id}")
        return ReportTask.model_validate(res["task"])

    def get_comparison_report(self, task: ReportTask) -> bytes:
        """
        Get the report for a specific task.
        :param task: The report task object containing the task ID.
        :return: The binary content of the comparison report.
        """
        return self._http.get_binary(
            f"{self._url}/alm/comparisonReports/"
            f"{task.result.target_revision_id}/{task.result.source_revision_id}"
        )

    @overload
    def create_comparison_summary(
        self,
        source_revision_id: str,
        source_model_id: str,
        target_revision_id: str,
        wait_for_completion: Literal[True] = True,
    ) -> SummaryReport: ...

    @overload
    def create_comparison_summary(
        self,
        source_revision_id: str,
        source_model_id: str,
        target_revision_id: str,
        wait_for_completion: Literal[False] = False,
    ) -> ReportTask: ...

    def create_comparison_summary(
        self,
        source_revision_id: str,
        source_model_id: str,
        target_revision_id: str,
        wait_for_completion: bool = True,
    ) -> ReportTask | SummaryReport:
        """
        Generate a comparison summary between two revisions.
        :param source_revision_id: The ID of the source revision.
        :param source_model_id: The ID of the source model.
        :param target_revision_id: The ID of the target revision.
        :param wait_for_completion: If True, the method will poll the task status and not return
               until the task is complete. If False, it will spawn the task and return immediately.
        :return: The created summary task or the summary report, if `wait_for_completion` is True.
        """
        payload = {
            "sourceRevisionId": source_revision_id,
            "sourceModelId": source_model_id,
            "targetRevisionId": target_revision_id,
        }
        res = self._http.post(f"{self._url}/alm/summaryReportTasks", json=payload)
        task = self.get_comparison_summary_task(res["task"]["taskId"])
        logger.info(
            f"Started Comparison Summary task '{task.id}' between Model '{source_model_id}' "
            f"(Revision '{source_revision_id}') and Model '{self._model_id}'."
        )
        if not wait_for_completion:
            return task
        task = self._http.poll_task(self.get_comparison_summary_task, task.id)
        if not task.result.successful:
            msg = f"Comparison Summary task {task.id} completed with errors: {task.result.error}."
            logger.error(msg)
            raise AnaplanActionError(msg)
        logger.info(f"Comparison Summary task {task.id} completed successfully.")
        return self.get_comparison_summary(task)

    def get_comparison_summary_task(self, task_id: str) -> ReportTask:
        """
        Get the task information for a comparison summary task.
        :param task_id: The ID of the comparison summary task.
        :return: The report task information.
        """
        res = self._http.get(f"{self._url}/alm/summaryReportTasks/{task_id}")
        return ReportTask.model_validate(res["task"])

    def get_comparison_summary(self, task: ReportTask) -> SummaryReport:
        """
        Get the comparison summary for a specific task.
        :param task: The summary task object containing the task ID.
        :return: The binary content of the comparison summary.
        """
        res = self._http.get(
            f"{self._url}/alm/summaryReports/"
            f"{task.result.target_revision_id}/{task.result.source_revision_id}"
        )
        return SummaryReport.model_validate(res["summaryReport"])



from anaplan_sdk import Client
from anaplan_sdk.models.scim import (
    NameInput,
    Remove,
    Replace,
    ReplaceUserInput,
    Resource,
    Schema,
    ServiceProviderConfig,
    User,
    field,
)


def test_get_service_provider_config(client: Client):
    conf = client.scim.get_service_provider_config()
    assert isinstance(conf, ServiceProviderConfig)


def test_get_resource_types(client: Client):
    res_types = client.scim.get_resource_types()
    assert isinstance(res_types, list)
    assert all(isinstance(r, Resource) for r in res_types)


def test_get_resource_schemas(client: Client):
    schemas = client.scim.get_resource_schemas()
    assert isinstance(schemas, list)
    assert all(isinstance(s, Schema) for s in schemas)


def test_get_users(client: Client):
    users = client.scim.get_users()
    assert isinstance(users, list)
    assert all(isinstance(u, User) for u in users)


def test_get_user_filtered(client: Client, name: str = "test.user@valantic.com"):
    users = client.scim.get_users(field("userName") == name)
    assert isinstance(users, list)
    assert all(isinstance(u, User) for u in users)
    assert all(u.user_name == name for u in users)


def test_get_user(client: Client, scim_user_id: str):
    user = client.scim.get_user(scim_user_id)
    assert isinstance(user, User)
    assert user.id == scim_user_id


def test_replace_user(client: Client, scim_user_id: str):
    user = client.scim.replace_user(
        scim_user_id,
        ReplaceUserInput(
            id=scim_user_id,
            name=NameInput(given_name="Test", family_name="User"),
            user_name="test.user@valantic.com",
        ),
    )
    assert isinstance(user, User)
    assert user.id == scim_user_id
    assert user.name.given_name == "Test"
    assert user.name.family_name == "User"


def test_create_update_user(client: Client, scim_user_id: str):
    user = client.scim.update_user(
        scim_user_id, [Replace(path="active", value=False), Remove(path="entitlements")]
    )
    assert isinstance(user, User)
    assert user.id == scim_user_id
    assert user.active is False
    assert user.entitlements == []



import logging
import os
from base64 import b64encode
from typing import Callable

import httpx

from ._oauth import _OAuthRequestFactory
from .exceptions import AnaplanException, InvalidCredentialsException, InvalidPrivateKeyException

logger = logging.getLogger("anaplan_sdk")


class _AnaplanAuth(httpx.Auth):
    requires_response_body = True

    def __init__(self, token: str | None = None):
        self._token: str = token or ""
        if not token:
            logger.info("Creating Authentication Token.")
            with httpx.Client(timeout=15.0) as client:
                self._parse_auth_response(client.send(self._build_auth_request()))

    def _build_auth_request(self) -> httpx.Request:
        raise NotImplementedError("Must be implemented in subclass.")

    def auth_flow(self, request):
        request.headers["Authorization"] = f"AnaplanAuthToken {self._token}"
        response = yield request
        if response.status_code == 401:
            logger.info("Token expired, refreshing.")
            auth_res = yield self._build_auth_request()
            self._parse_auth_response(auth_res)
            request.headers["Authorization"] = f"AnaplanAuthToken {self._token}"
            yield request

    def _parse_auth_response(self, response: httpx.Response) -> None:
        if response.status_code == 401:
            raise InvalidCredentialsException
        if not response.is_success:
            raise AnaplanException(f"Authentication failed: {response.status_code} {response.text}")
        self._token = response.json()["tokenInfo"]["tokenValue"]


class _StaticTokenAuth(httpx.Auth):
    def __init__(self, token: str):
        self._token = token

    def auth_flow(self, request):
        request.headers["Authorization"] = f"AnaplanAuthToken {self._token}"
        response = yield request
        if response.status_code == 401:
            raise InvalidCredentialsException("Token is invalid or expired.")


class _AnaplanBasicAuth(_AnaplanAuth):
    def __init__(self, user_email: str, password: str, token: str | None = None):
        self.user_email = user_email
        self.password = password
        super().__init__(token)

    def _build_auth_request(self) -> httpx.Request:
        cred = b64encode(f"{self.user_email}:{self.password}".encode()).decode()
        return httpx.Request(
            method="post",
            url="https://auth.anaplan.com/token/authenticate",
            headers={"Authorization": f"Basic {cred}"},
        )


class _AnaplanCertAuth(_AnaplanAuth):
    requires_request_body = True

    def __init__(
        self,
        certificate: str | bytes,
        private_key: str | bytes,
        private_key_password: str | bytes | None = None,
        token: str | None = None,
    ):
        self.__set_certificate(certificate)
        self.__set_private_key(private_key, private_key_password)
        super().__init__(token)

    def _build_auth_request(self) -> httpx.Request:
        encoded_cert, encoded_string, encoded_signed_string = self._prep_credentials()
        return httpx.Request(
            method="post",
            url="https://auth.anaplan.com/token/authenticate",
            headers={
                "Authorization": f"CACertificate {encoded_cert}",
                "Content-Type": "application/json",
            },
            json={"encodedData": encoded_string, "encodedSignedData": encoded_signed_string},
        )

    def _prep_credentials(self) -> tuple[str, str, str]:
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15

        message = os.urandom(150)
        return (
            b64encode(self._certificate).decode(),
            b64encode(message).decode(),
            b64encode(self._private_key.sign(message, PKCS1v15(), hashes.SHA512())).decode(),
        )

    def __set_certificate(self, certificate: str | bytes) -> None:
        if isinstance(certificate, str):
            if os.path.isfile(certificate):
                with open(certificate, "rb") as f:
                    self._certificate = f.read()
            else:
                self._certificate = certificate.encode()
        else:
            self._certificate = certificate

    def __set_private_key(
        self, private_key: str | bytes, private_key_password: str | bytes
    ) -> None:
        try:
            from cryptography.exceptions import InvalidKey, UnsupportedAlgorithm
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import serialization
            from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
        except ImportError as e:
            raise AnaplanException(
                "cryptography is not available. Please install anaplan-sdk with the cert extra "
                "`pip install anaplan-sdk[cert]` or install cryptography separately."
            ) from e
        try:
            if isinstance(private_key, str):
                if os.path.isfile(private_key):
                    with open(private_key, "rb") as f:
                        data = f.read()
                else:
                    data = private_key.encode()
            else:
                data = private_key
            password = (
                private_key_password.encode()
                if isinstance(private_key_password, str)
                else private_key_password
            )
            self._private_key: RSAPrivateKey = serialization.load_pem_private_key(
                data, password, backend=default_backend()
            )
        except (IOError, InvalidKey, UnsupportedAlgorithm) as error:
            raise InvalidPrivateKeyException from error


class AnaplanLocalOAuth(_AnaplanAuth):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        token: dict[str, str] | None = None,
        persist_token: bool = False,
        authorization_url: str = "https://us1a.app.anaplan.com/auth/prelogin",
        token_url: str = "https://us1a.app.anaplan.com/oauth/token",
        validation_url: str = "https://auth.anaplan.com/token/validate",
        scope: str = "openid profile email offline_access",
        state_generator: Callable[[], str] | None = None,
    ):
        """
        Initializes the AnaplanLocalOAuth class for OAuth2 authentication using the
        Authorization Code Flow. This is a utility class for local development and requires user
        interaction. For Web Applications and other scenarios, refer to `Oauth` or `AsyncOauth`.
        This class will refresh the access token automatically when it expires.
        :param client_id: The client ID of your Anaplan Oauth 2.0 application. This Application
               must be an Authorization Code Grant application.
        :param client_secret: The client secret of your Anaplan Oauth 2.0 application.
        :param redirect_uri: The URL to which the user will be redirected after authorizing the
               application.
        :param token: The OAuth token dictionary containing at least the `access_token` and
               `refresh_token`. If not provided, the user will be prompted to interactive
               authorize the application, if `persist_token` is set to False or no valid refresh
               token is found in the keyring.
        :param persist_token: If set to True, the refresh token will be stored in the system's
               keyring, allowing the application to use the same refresh token across multiple
               runs. If set to False, the user will be prompted to authorize the application each
               time. This requires the `keyring` extra to be installed. If a valid refresh token
               is found in the keyring, it will be used instead of the given `token` parameter.
        :param authorization_url: The URL to which the user will be redirected to authorize the
               application. Defaults to the Anaplan Prelogin Page, where the user can select the
               login method.
        :param token_url: The URL to post the authorization code to in order to fetch the access
               token.
        :param validation_url: The URL to validate the access token.
        :param scope: The scope of the access request.
        :param state_generator: A callable that generates a random state string. You can optionally
               pass this if you need to customize the state generation logic. If not provided,
               the state will be generated by `oauthlib`.
        """
        self._oauth_token = token or {}
        self._service_name = "anaplan_sdk"

        if persist_token:
            try:
                import keyring

                stored = keyring.get_password(self._service_name, self._service_name)
                if stored:
                    logger.info("Using persisted OAuth refresh token.")
                    self._oauth_token = {"refresh_token": stored}
                    self._token = ""  # Set to blank to trigger the super().__init__ auth request.
            except ImportError as e:
                raise AnaplanException(
                    "keyring is not available. Please install anaplan-sdk with the keyring extra "
                    "`pip install anaplan-sdk[keyring]` or install keyring separately."
                ) from e
        self._persist_token = persist_token
        self._oauth = _OAuthRequestFactory(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            scope=scope,
            authorization_url=authorization_url,
            token_url=token_url,
            validation_url=validation_url,
            state_generator=state_generator,
        )
        if not self._oauth_token:
            self.__auth_code_flow()
        super().__init__(self._token)

    @property
    def token(self) -> dict[str, str]:
        """
        Returns the current token dictionary. You can safely use the `access_token`, but if you
        must not use the `refresh_token` outside of this class, if you expect to use this instance
        further. If you do use the `refresh_token` outside of this class, this will error on the
        next attempt to refresh the token, as the `refresh_token` can only be used once.
        """
        return self._oauth_token

    def _build_auth_request(self) -> httpx.Request:
        return self._oauth.refresh_token_request(self._oauth_token["refresh_token"])

    def _parse_auth_response(self, response: httpx.Response) -> None:
        if response.status_code == 401:
            raise InvalidCredentialsException
        if not response.is_success:
            raise AnaplanException(f"Authentication failed: {response.status_code} {response.text}")
        self._oauth_token = response.json()
        if self._persist_token:
            import keyring

            keyring.set_password(
                self._service_name, self._service_name, self._oauth_token["refresh_token"]
            )
        self._token: str = self._oauth_token["access_token"]

    def __auth_code_flow(self):
        from oauthlib.oauth2 import OAuth2Error

        try:
            logger.info("Creating Authentication Token with OAuth2 Authorization Code Flow.")
            url, _ = self._oauth.authorization_url()
            authorization_response = input(
                f"Please go to {url} and authorize the app.\n"
                "Then paste the entire redirect URL here: "
            )
            with httpx.Client() as client:
                res = client.send(self._oauth.token_request(authorization_response))
            self._parse_auth_response(res)
        except (httpx.HTTPError, ValueError, TypeError, OAuth2Error) as error:
            raise InvalidCredentialsException("Error during OAuth2 authorization flow.") from error


class AnaplanRefreshTokenAuth(_AnaplanAuth):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        token: dict[str, str],
        token_url: str = "https://us1a.app.anaplan.com/oauth/token",
    ):
        """
        This class is a utility class for long-lived `Client` or `AsyncClient` instances that use
        OAuth. This class will use the `access_token` until the first request fails with a 401
        Unauthorized error, at which point it will attempt to refresh the `access_token` using the
        `refresh_token`. If the refresh fails, it will raise an `InvalidCredentialsException`. The
        `expires_in` field in the token dictionary is not considered. Manipulating any of the
        fields in the token dictionary is not recommended and will likely have no effect.

        **For its entire lifetime, you are ceding control of the token to this class.**
        You must not use the same token simultaneously in multiple instances of this class or
        outside of it, as this may lead to unexpected behavior when e.g. the refresh token is
        used, which can only happen once and will lead to errors when attempting to use the same
        refresh token again elsewhere.

        If you need the token back before this instance is destroyed, you can use the `token`
        method.

        :param client_id: The client ID of your Anaplan Oauth 2.0 application. This Application
               must be an Authorization Code Grant application.
        :param client_secret: The client secret of your Anaplan Oauth 2.0 application.
        :param redirect_uri: The URL to which the user will be redirected after authorizing the
               application.
        :param token: The OAuth token dictionary containing at least the `access_token` and
               `refresh_token`.
        :param token_url: The URL to post the refresh token request to in order to fetch the access
               token.
        """
        if not isinstance(token, dict) or not all(
            key in token for key in ("access_token", "refresh_token")
        ):
            raise ValueError(
                "The token must at least contain 'access_token' and 'refresh_token' keys."
            )
        self._oauth_token = token
        self._oauth = _OAuthRequestFactory(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            token_url=token_url,
        )
        super().__init__(self._oauth_token["access_token"])

    @property
    def token(self) -> dict[str, str]:
        """
        Returns the current OAuth token. You can safely use the `access_token`, but you
        must not use the `refresh_token` outside of this class, if you expect to use this instance
        further. If you do use the `refresh_token` outside of this class, this will error on the
        next attempt to refresh the token, as the `refresh_token` can only be used once.
        """
        return self._oauth_token

    def _build_auth_request(self) -> httpx.Request:
        return self._oauth.refresh_token_request(self._oauth_token["refresh_token"])

    def _parse_auth_response(self, response: httpx.Response) -> None:
        if response.status_code == 401:
            raise InvalidCredentialsException
        if not response.is_success:
            raise AnaplanException(f"Authentication failed: {response.status_code} {response.text}")
        self._oauth_token = response.json()
        self._token: str = self._oauth_token["access_token"]


def _create_auth(
    user_email: str | None = None,
    password: str | None = None,
    certificate: str | bytes | None = None,
    private_key: str | bytes | None = None,
    private_key_password: str | bytes | None = None,
    token: str | None = None,
) -> httpx.Auth:
    if certificate and private_key:
        return _AnaplanCertAuth(certificate, private_key, private_key_password, token)
    if user_email and password:
        return _AnaplanBasicAuth(user_email, password, token)
    if token:
        return _StaticTokenAuth(token)
    raise ValueError(
        "No valid authentication parameters provided. Please provide either:\n"
        "- `user_email` and `password`, or\n"
        "- `certificate` and `private_key`\n"
    )



from pydantic import Field

from ._base import AnaplanModel
from ._bulk import TaskSummary


class Revision(AnaplanModel):
    id: str = Field(description="The unique identifier of this revision.")
    name: str = Field(description="The name of this revision.")
    description: str | None = Field(
        None, description="The description of this revision. Not always present."
    )
    created_on: str = Field(description="The creation date of this revision in ISO format.")
    created_by: str = Field(
        description="The unique identifier of the user who created this revision."
    )
    creation_method: str = Field(description="The creation method of this revision.")
    applied_on: str = Field(description="The application date of this revision in ISO format.")
    applied_by: str = Field(
        description="The unique identifier of the user who applied this revision."
    )


class ModelRevision(AnaplanModel):
    id: str = Field(
        validation_alias="modelId",
        description="The unique identifier of the model this revision belongs to.",
    )
    name: str = Field(
        default="",
        validation_alias="modelName",
        description=(
            "The name of the model this revision belongs to. This can be an empty string, when the "
            "calling user does not have access to the model, but is workspace admin in the "
            "workspace."
        ),
    )
    workspace_id: str | None = Field(
        None, description="The unique identifier of the workspace this revision belongs to."
    )
    applied_by: str = Field(
        description="The unique identifier of the user who applied this revision."
    )
    applied_on: str = Field(description="The application date of this revision in ISO format.")
    applied_method: str = Field(description="The application method of this revision.")
    deleted: bool | None = Field(
        None,
        validation_alias="modelDeleted",
        description="Whether the model has been deleted or not.",
    )


class SyncTaskResult(AnaplanModel):
    source_revision_id: str = Field(description="The ID of the source revision.")
    target_revision_id: str = Field(description="The ID of the target revision.")
    successful: bool = Field(description="Whether the sync task was successful or not.")


class SyncTask(TaskSummary):
    current_step: str = Field(description="The current step of the sync task.")
    result: SyncTaskResult | None = Field(None, description="The result of the sync task.")


class ReportTaskResult(SyncTaskResult):
    report_file_url: str = Field(
        description="The URL of the report file generated by the sync task."
    )


class ReportTaskError(AnaplanModel):
    title: str = Field(description="The title of the error.")
    message: str = Field(validation_alias="messageText", description="The message of the error.")


class ReportTaskFailureResult(AnaplanModel):
    successful: bool = Field(description="Whether the sync task was successful or not.")
    error: ReportTaskError = Field(description="The error that occurred during the sync task.")


class ReportTask(SyncTask):
    result: ReportTaskResult | ReportTaskFailureResult | None = Field(
        None, description="The result of the comparison report task, including the report file URL."
    )


class SummaryTotals(AnaplanModel):
    modified: int = Field(0, description="The number of modified items.")
    deleted: int = Field(0, description="The number of deleted items.")
    created: int = Field(0, description="The number of created items.")


class SummaryDifferences(AnaplanModel):
    line_items: SummaryTotals = Field(
        SummaryTotals(modified=0, deleted=0, created=0), description="Changes in line items."
    )
    roles_contents: SummaryTotals = Field(
        SummaryTotals(modified=0, deleted=0, created=0), description="Changes in roles contents."
    )
    lists: SummaryTotals = Field(
        SummaryTotals(modified=0, deleted=0, created=0), description="Changes in lists."
    )
    modules: SummaryTotals = Field(
        SummaryTotals(modified=0, deleted=0, created=0), description="Changes in modules."
    )


class SummaryReport(AnaplanModel):
    target_revision_id: str = Field(description="The ID of the target revision.")
    source_revision_id: str = Field(description="The ID of the source revision.")
    totals: SummaryTotals = Field(description="The total counts of changes.")
    differences: SummaryDifferences = Field(
        description="The detailed breakdown of changes by category."
    )



from datetime import datetime
from typing import Literal

from pydantic import Field, field_validator

from ._base import AnaplanModel
from ._bulk import TaskResultDetail

ConnectionType = Literal["AmazonS3", "AzureBlob", "GoogleBigQuery"]
IntegrationType = Literal[
    "AmazonS3ToAnaplan",
    "AzureBlobToAnaplan",
    "GoogleBigQueryToAnaplan",
    "AnaplanToAmazonS3",
    "AnaplanToAzureBlob",
    "AnaplanToGoogleBigQuery",
]

Timezone = Literal[
    "Etc/GMT+12",
    "US/Samoa",
    "Pacific/Honolulu",
    "Pacific/Marquesas",
    "US/Aleutian",
    "America/Anchorage",
    "America/Los_Angeles",
    "America/Denver",
    "America/Chicago",
    "America/New_York",
    "America/Sao_Paulo",
    "Canada/Newfoundland",
    "America/Nuuk",
    "Atlantic/Cape_Verde",
    "Greenwich",
    "Europe/London",
    "Europe/Paris",
    "Asia/Tel_Aviv",
    "Europe/Moscow",
    "Asia/Dubai",
    "Asia/Kabul",
    "Asia/Karachi",
    "Asia/Kolkata",
    "Asia/Kathmandu",
    "Asia/Dhaka",
    "Asia/Rangoon",
    "Asia/Jakarta",
    "Asia/Hong_Kong",
    "Australia/Eucla",
    "Asia/Tokyo",
    "Australia/Adelaide",
    "Australia/Brisbane",
    "Australia/Melbourne",
    "Australia/Perth",
    "Australia/Lord_Howe",
    "Pacific/Norfolk",
    "Pacific/Auckland",
    "Pacific/Chatham",
]


class _VersionedBaseModel(AnaplanModel):
    creation_date: datetime = Field(description="The initial creation date.")
    modification_date: datetime = Field(
        description=(
            "The last modification date. If never modified, this is equal to creation_date."
        )
    )
    created_by: str = Field(description="The user who created this.")
    modified_by: str | None = Field(description="The user who last modified this.")


class BaseConnectionInput(AnaplanModel):
    workspace_id: str | None = Field(
        default=None,
        description=(
            "If you are a restricted integration user, add the Workspace ID to which you have "
            "access in the payload."
        ),
    )


class AzureBlobConnectionInfo(AnaplanModel):
    name: str = Field(description="The name of the Azure Blob connection.")
    storage_account_name: str = Field(description="The name of the Azure Storage account.")
    container_name: str = Field(description="The name of the Azure Blob container.")


class AzureBlobConnectionInput(AzureBlobConnectionInfo, BaseConnectionInput):
    sas_token: str = Field(
        description=(
            "The SAS token for the Azure Blob connection. Must be created on the container "
            "directly and not on any child blobs and have at least 'Read' and 'List' permissions."
        )
    )


class AmazonS3ConnectionInfo(AnaplanModel):
    name: str = Field(description="The name of the Amazon S3 connection.")
    bucket_name: str = Field(description="The name of the Amazon S3 bucket.")


class AmazonS3ConnectionInput(AmazonS3ConnectionInfo, BaseConnectionInput):
    access_key_id: str = Field(description="The access key ID for the Amazon S3 connection.")
    secret_access_key: str = Field(
        description="The secret access key for the Amazon S3 connection."
    )


class GoogleBigQueryConnectionInfo(AnaplanModel):
    name: str = Field(description="The name of the Google BigQuery connection.")
    dataset: str = Field(description="The ID of the Google BigQuery dataset.")


class GoogleServiceAccountJson(AnaplanModel):
    type: str = Field(description="The type of the service account.")
    project_id: str = Field(description="The project ID of the service account.")
    private_key_id: str = Field(description="The private key ID of the service account.")
    private_key: str = Field(description="The private key of the service account.")
    client_email: str = Field(description="The client email of the service account.")
    client_id: str = Field(description="The client ID of the service account.")
    auth_uri: str = Field(description="The authentication URI of the service account.")
    token_uri: str = Field(description="The token URI of the service account.")
    auth_provider_x509_cert_url: str = Field(
        description="The authentication provider's X.509 certificate URL."
    )
    client_x509_cert_url: str = Field(description="The client's X.509 certificate URL.")


class GoogleBigQueryConnectionInput(GoogleBigQueryConnectionInfo, BaseConnectionInput):
    serviceAccountKey: GoogleServiceAccountJson = Field(
        description="The service account JSON for the Google BigQuery connection."
    )


ConnectionBody = AzureBlobConnectionInput | AmazonS3ConnectionInput | GoogleBigQueryConnectionInput


class ConnectionInput(AnaplanModel):
    type: ConnectionType = Field(description="The type of this connection.")
    body: ConnectionBody = Field(description="Connection information.")


class Connection(_VersionedBaseModel):
    id: str = Field(
        validation_alias="connectionId", description="The unique identifier of this connection."
    )
    connection_type: ConnectionType = Field(description="The type of this connection.")
    body: AzureBlobConnectionInfo | AmazonS3ConnectionInfo | GoogleBigQueryConnectionInfo = Field(
        description="Connection information."
    )
    status: int = Field(
        description=(
            "The status of this connection. 1 indicates a valid connection, 0 indicates an invalid "
            "connection."
        )
    )
    integration_error_code: str | None = Field(
        description="The error code of the connection, if any."
    )
    workspace_id: str | None = Field(
        description="The workspace that was given when creating this connection."
    )


class LatestRun(AnaplanModel):
    triggered_by: str = Field(description="The user who triggered this run.")
    start_date: datetime = Field(description="The start timestamp of this run.")
    end_date: datetime | None = Field(
        default=None,
        description=(
            "The end timestamp of this run. This can be None, if the integration is currently "
            "running."
        ),
    )
    success: bool = Field(description="Whether this run was successful.")
    message: str = Field(description="Result message of this run.")
    execution_error_code: int | None = Field(default=None, description="Error code if run failed.")
    trigger_source: str | None = Field(default=None, description="Source that triggered the run.")


class ScheduleBase(AnaplanModel):
    name: str = Field(description="Name of the schedule.")
    type: Literal[
        "hourly", "daily", "weekly", "monthly_specific_day", "monthly_relative_weekday"
    ] = Field(description="Trigger Frequency")
    timezone: Timezone = Field(description="Timezone for the schedule.")


class ScheduleInput(ScheduleBase):
    time: str | None = Field(default=None, description="Time for scheduled runs in HH:mm format.")
    from_time: str | None = Field(
        default=None, description="Time for scheduled runs in HH:mm format, if type is hourly."
    )
    to_time: str | None = Field(
        default=None, description="Time for scheduled runs in HH:mm format, if type is hourly."
    )
    days_of_week: list[int] = Field(description="Days of week when schedule is active.")
    start_date: str = Field(
        description=(
            "Start date for the schedule in YYYY-MM-DD format. Must be in the Future, i.e. current "
            "day, if the `time` is greater than the current time or any future date. "
        )
    )
    end_date: str | None = Field(
        default=None,
        description=(
            "End date for the schedule in YYYY-MM-DD format. Must be in the Future, i.e. current "
            "day, if the `time` is greater than the current time or any future date. Can also be "
            "omitted to create a schedule that runs indefinitely."
        ),
    )


class Schedule(ScheduleBase):
    time: str | None = Field(default=None, description="Time for scheduled runs in HH:mm format.")
    to_time: str | None = Field(default=None, description="End time for scheduled runs.")
    from_time: str | None = Field(default=None, description="Start time for scheduled runs.")
    start_date: datetime = Field(description="Start date of the schedule.")
    end_date: datetime | None = Field(default=None, description="End date of the schedule, if set.")
    days_of_week: list[int] = Field(default=[], description="Days of week when schedule is active.")
    repeat_every: int | None = Field(default=None, description="Frequency of repetition.")
    status: str = Field(description="Current status of the schedule.")


class _BaseIntegration(_VersionedBaseModel):
    name: str = Field(description="The name of this integration.")
    created_by: str = Field(description="The user who created this integration.")
    notification_id: str | None = Field(
        default=None, description="The ID of the associated notification configuration, if any."
    )
    latest_run: LatestRun | None = Field(
        default=None, description="Details about the latest execution, if any."
    )

    @field_validator("latest_run", mode="before")
    @classmethod
    def _empty_source_is_none(cls, inp: dict):
        return inp if inp else None


class Integration(_BaseIntegration):
    id: str = Field(
        validation_alias="integrationId", description="The unique identifier of this integration."
    )
    integration_type: Literal["Import", "Export", "Process"] = Field(
        description="The type of this integration."
    )
    model_id: str = Field(description="The ID of the model this integration belongs to.")
    workspace_id: str = Field(description="The ID of the workspace this integration belongs to.")
    nux_visible: bool = Field(description="Whether this integration is visible in the UI.")
    process_id: str | None = Field(None, description="The ID of the process (for Process type).")
    schedule: Schedule | None = Field(
        default=None, description="Schedule configuration if defined."
    )


class SingleIntegration(Integration):
    integration_type: None = Field(
        default=None,
        description=(
            "Sentinel for erroneous implementation of the Anaplan API. This field is not provided "
            "when getting an individual integration by Id."
        ),
    )


class AnaplanSource(AnaplanModel):
    type: Literal["Anaplan"] = Field(
        default="Anaplan", description="Literal signifying this is an Anaplan source."
    )
    action_id: int = Field(
        description=(
            "The ID of the action to be used as a source. This can be a process, or export."
        )
    )


class FileSource(AnaplanModel):
    connection_id: str = Field(description="The unique identifier of the connection.")
    type: Literal["AmazonS3", "AzureBlob"] = Field(description="The type of this connection.")
    file: str = Field(description="The file path relative to the root of the connection.")


class FileTarget(FileSource):
    connection_id: str = Field(description="The unique identifier of the connection.")
    type: Literal["AmazonS3", "AzureBlob"] = Field(description="The type of this connection.")
    overwrite: bool = Field(default=True, description="Whether to overwrite the file if it exists.")


class TableSource(AnaplanModel):
    type: Literal["GoogleBigQuery"] = Field(
        default="GoogleBigQuery", description="The type of this connection."
    )
    connection_id: str = Field(description="The unique identifier of the connection.")
    table: str = Field(description="The table name in the BigQuery dataset in the connection.")


class TableTarget(TableSource):
    overwrite: bool = Field(
        default=False, description="Whether to overwrite the table if it exists."
    )


class AnaplanTarget(AnaplanModel):
    type: Literal["Anaplan"] = Field(
        default="Anaplan", description="Literal signifying this is an Anaplan target."
    )
    action_id: int = Field(
        description=(
            "The ID of the action to be used as a target. This can be a process, or import."
        )
    )
    file_id: int = Field(description="The ID of the file to be used as a target.")


class IntegrationJobInput(AnaplanModel):
    type: IntegrationType = Field(description="The type of this integration.")
    sources: list[AnaplanSource | FileSource | TableSource] = Field(
        description="The sources of this integration."
    )
    targets: list[AnaplanTarget | FileTarget | TableTarget] = Field(
        description="The targets of this integration."
    )


class IntegrationProcessInput(AnaplanModel):
    name: str = Field(description="The name of this integration process.")
    version: Literal["2.0"] = Field(default="2.0", description="The version of this integration.")
    workspace_id: str = Field(description="The ID of the workspace this integration belongs to.")
    model_id: str = Field(description="The ID of the model this integration belongs to.")
    process_id: int = Field(description="The ID of the process this integration belongs to.")


class IntegrationInput(AnaplanModel):
    name: str = Field(description="The name of this integration.")
    version: Literal["2.0"] = Field(default="2.0", description="The version of this integration.")
    workspace_id: str = Field(description="The ID of the workspace this integration belongs to.")
    model_id: str = Field(description="The ID of the model this integration belongs to.")
    process_id: int | None = Field(
        default=None,
        description=(
            "If given, an integration process will be created, instead of an Import or Export"
        ),
    )
    nux_visible: bool = Field(
        default=False, description="Whether this integration is visible in the UI."
    )
    jobs: list[IntegrationJobInput] = Field(
        description="The jobs in this integration.", min_length=1
    )


class RunSummary(AnaplanModel):
    id: str = Field(description="The unique identifier of this run.")
    triggered_by: str
    last_run: datetime = Field(description="Last Run timestamp.")
    start_date: datetime = Field(description="Start timestamp.")
    end_date: datetime | None = Field(
        default=None,
        description=(
            "The end timestamp of this run. This can be None, if the integration is currently "
            "running."
        ),
    )
    success: bool = Field(description="Whether this run was successful.")
    message: str = Field(description="Result message of this run.")
    execution_error_code: int | None = Field(default=None, description="Error code if run failed.")
    trace_id: str = Field(description="The trace ID for this run.")
    trigger_source: Literal["manual", "scheduled"] = Field(
        description="Source that triggered the run."
    )


class RunStatus(AnaplanModel):
    id: str = Field(description="The unique identifier of this run.")
    integration_id: str = Field(description="The ID of the integration this run belongs to.")
    trace_id: str = Field(description="The trace ID for this run.")
    start_date: datetime = Field(description="The start timestamp of this run.")
    end_date: datetime | None = Field(
        default=None,
        description=(
            "The end timestamp of this run. This can be None, if the integration is currently "
            "running."
        ),
    )
    success: bool = Field(description="Whether this run was successful.")
    message: str = Field(description="Result message of this run.")
    execution_error_code: int | None = Field(default=None, description="Error code if run failed.")
    flow_group_id: str | None = Field(default=None, description="The ID of the flow group, if any.")
    trigger_source: Literal["manual", "scheduled"] = Field(
        description="Source that triggered the run."
    )


class ErrorSummary(AnaplanModel):
    local_message_text: str = Field(description="Error message text.")


class ErrorMessage(AnaplanModel):
    error_message: list[TaskResultDetail | ErrorSummary]
    action_id: str = Field(description="The ID of the action that failed.")
    action_name: str = Field(description="The name of the action that failed.")
    failure_dump_generated: bool = Field(description="Whether a failure dump was generated.")


class RunError(AnaplanModel):
    task_id: str = Field(description="The Task ID of the invoked Anaplan Action.")
    error_messages: list[ErrorMessage] = Field(description="The error messages of the run.")


class NotificationUser(AnaplanModel):
    user_guid: str = Field(description="The unique identifier of the user.")
    first_name: str = Field(description="The user's first name.")
    last_name: str = Field(description="The user's last name.")


class NotificationItem(AnaplanModel):
    type: Literal["success", "partial_failure", "full_failure"] = Field(
        description="The type of notification event that triggers notifications."
    )
    users: list[NotificationUser] = Field(
        description="The list of users who will receive this notification."
    )


class Notification(AnaplanModel):
    config: list[NotificationItem] = Field(
        description="The configuration for different notification types."
    )


class NotificationConfig(AnaplanModel):
    notification_id: str = Field(
        description="The unique identifier of this notification configuration."
    )
    integration_ids: list[str] = Field(
        description="The IDs of the integrations associated with this notification."
    )
    channels: list[Literal["email", "in_app"]] = Field(
        description="The channels through which notifications will be sent."
    )
    notifications: Notification = Field(
        description="The detailed notification configuration settings."
    )


class NotificationItemInput(AnaplanModel):
    type: Literal["success", "partial_failure", "full_failure"] = Field(
        description="The type of notification event that triggers notifications."
    )
    users: list[str] = Field(
        description=(
            "The list of user IDs who will receive this notification. Must not be empty. If you "
            "want nobody to receive notifications for this type, omit the entire config item. If "
            "you want to override an existing list of users with an empty one, you must delete the "
            "notification configuration and create a new one."
        ),
        min_length=1,
        max_length=5,
    )


class NotificationConfigInput(AnaplanModel):
    config: list[NotificationItemInput] = Field(
        description="The configuration for different notification types."
    )


class NotificationInput(AnaplanModel):
    integration_ids: list[str] = Field(
        description="The IDs of the integrations associated with this notification."
    )
    channels: list[Literal["email", "in_app"]] = Field(
        description="The channels through which notifications will be sent."
    )
    notifications: NotificationConfigInput = Field(
        description="The detailed notification configuration settings."
    )



from ._async_clients import AsyncClient
from ._auth import AnaplanLocalOAuth, AnaplanRefreshTokenAuth
from ._clients import Client
from ._oauth import AsyncOauth, Oauth
from .models.scim import field

__all__ = [
    "AsyncClient",
    "Client",
    "AnaplanLocalOAuth",
    "AnaplanRefreshTokenAuth",
    "AsyncOauth",
    "Oauth",
    "models",
    "exceptions",
    "field",
]



from anaplan_sdk import field


def test_active():
    assert str(field("active")) == "active eq true"


def test_not_active():
    assert str(~field("active")) == "active eq false"


def test_name_truthy():
    assert str(field("userName")) == "userName pr"


def test_name_falsy():
    assert str(~field("userName")) == "userName eq null"


def test_eq():
    assert str(field("id") == "123") == 'id eq "123"'


def test_ne():
    assert str(field("id") != "123") == 'id ne "123"'


def test_gt():
    assert str(field("name.givenName") > "zzz") == 'name.givenName gt "zzz"'


def test_ge():
    assert str(field("name.givenName") >= "zzz") == 'name.givenName ge "zzz"'


def test_lt():
    assert str(field("name.givenName") < "zzz") == 'name.givenName lt "zzz"'


def test_le():
    assert str(field("name.givenName") <= "zzz") == 'name.givenName le "zzz"'


def test_and():
    assert str(field("active") & (field("id") == "123")) == 'active eq true and id eq "123"'


def test_or():
    assert str(field("active") | (field("id") == "123")) == 'active eq true or id eq "123"'


def test_op_order():
    expr = (field("active") & (field("id") == "123")) | ~field("userName")
    assert str(expr) == '(active eq true and id eq "123") or userName eq null'


def test_compose():
    part = field("active") & field("userName") & (field("name.givenName") > "Thomas")
    assert str(part) == 'active eq true and userName pr and name.givenName gt "Thomas"'
    predicate = (field("userName") == "test.user@valantic.com") | (
        part | (~field("active") & (field("name.givenName") != "Thomas"))
    )
    assert str(predicate) == (
        'userName eq "test.user@valantic.com" or ((active eq true and userName pr and '
        'name.givenName gt "Thomas") or (active eq false and name.givenName ne "Thomas"))'
    )



import logging
from typing import Any

from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import construct_payload
from anaplan_sdk.models.flows import Flow, FlowInput, FlowSummary

logger = logging.getLogger("anaplan_sdk")


class _FlowClient:
    def __init__(self, http: _HttpService) -> None:
        self._http = http
        self._url = "https://api.cloudworks.anaplan.com/2/0/integrationflows"

    def get_flows(self, current_user_only: bool = False) -> list[FlowSummary]:
        """
        List all flows in CloudWorks.
        :param current_user_only: Filters the flows to only those created by the current user.
        :return: A list of FlowSummaries.
        """
        params = {"myIntegrations": 1 if current_user_only else 0}
        return [
            FlowSummary.model_validate(e)
            for e in self._http.get_paginated(self._url, "integrationFlows", params=params)
        ]

    def get_flow(self, flow_id: str) -> Flow:
        """
        Get a flow by its ID. This returns the full flow object, including the contained steps and
        continuation behavior.
        :param flow_id: The ID of the flow to get.
        :return: The Flow object.
        """
        return Flow.model_validate((self._http.get(f"{self._url}/{flow_id}"))["integrationFlow"])

    def run_flow(self, flow_id: str, only_steps: list[str] = None) -> str:
        """
        Run a flow by its ID. Make sure that neither the flow nor any of its contained are running.
        If this is the case, the task will error. Anaplan neither schedules these tasks nor can it
        handle concurrent executions.
        :param flow_id: The ID of the flow to run.
        :param only_steps: A list of step IDs to run. If not provided, only these will be run.
        :return: The ID of the run.
        """
        url = f"{self._url}/{flow_id}/run"
        res = (
            self._http.post(url, json={"stepsToRun": only_steps})
            if only_steps
            else self._http.post_empty(url)
        )
        run_id = res["run"]["id"]
        logger.info(f"Started flow run '{run_id}' for flow '{flow_id}'.")
        return run_id

    def create_flow(self, flow: FlowInput | dict[str, Any]) -> str:
        """
        Create a new flow in CloudWorks. Be careful not to omit the `depends_on` field. Anaplan
        will accept these values, but an invalid, corrupted flow will be created, as all Flows must
        have at least 2 Steps, and they must always be sequential
        :param flow: The flow to create. This can be a FlowInput object or a dictionary.
        :return: The ID of the created flow.
        """
        res = self._http.post(self._url, json=construct_payload(FlowInput, flow))
        flow_id = res["integrationFlow"]["integrationFlowId"]
        logger.info(f"Created flow '{flow_id}'.")
        return flow_id

    def update_flow(self, flow_id: str, flow: FlowInput | dict[str, Any]) -> None:
        """
        Update a flow in CloudWorks. You must provide the full flow object, partial updates are not
        supported.
        :param flow_id: The ID of the flow to update.
        :param flow: The flow to update. This can be a FlowInput object or a dictionary.
        """
        self._http.put(f"{self._url}/{flow_id}", json=construct_payload(FlowInput, flow))
        logger.info(f"Updated flow '{flow_id}'.")

    def delete_flow(self, flow_id: str) -> None:
        """
        Delete a flow in CloudWorks. This will not delete its contained steps. This will fail if
        the flow is running or if it has any running steps.
        :param flow_id: The ID of the flow to delete.
        """
        self._http.delete(f"{self._url}/{flow_id}")
        logger.info(f"Deleted flow '{flow_id}'.")



import logging
from concurrent.futures import ThreadPoolExecutor
from itertools import chain
from typing import Any, Literal, overload

from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import (
    parse_calendar_response,
    parse_insertion_response,
    sort_params,
    validate_dimension_id,
)
from anaplan_sdk.exceptions import InvalidIdentifierException
from anaplan_sdk.models import (
    CurrentPeriod,
    Dimension,
    DimensionWithCode,
    FiscalYear,
    InsertionResult,
    LineItem,
    List,
    ListDeletionResult,
    ListItem,
    ListMetadata,
    Model,
    ModelCalendar,
    ModelStatus,
    Module,
    View,
    ViewInfo,
)

SortBy = Literal["id", "name"] | None

logger = logging.getLogger("anaplan_sdk")


class _TransactionalClient:
    def __init__(self, http: _HttpService, model_id: str) -> None:
        self._http = http
        self._url = f"https://api.anaplan.com/2/0/models/{model_id}"
        self._model_id = model_id

    def get_model_details(self) -> Model:
        """
        Retrieves the Model details for the current Model.
        :return: The Model details.
        """
        return Model.model_validate(
            self._http.get(self._url, params={"modelDetails": "true"})["model"]
        )

    def get_model_status(self) -> ModelStatus:
        """
        Gets the current status of the Model.
        :return: The current status of the Mode.
        """
        return ModelStatus.model_validate(
            self._http.get(f"{self._url}/status").get("requestStatus")
        )

    def wake_model(self) -> None:
        """Wake up the current model."""
        self._http.post_empty(f"{self._url}/open", headers={"Content-Type": "application/text"})
        logger.info(f"Woke up model '{self._model_id}'.")

    def close_model(self) -> None:
        """Close the current model."""
        self._http.post_empty(f"{self._url}/close", headers={"Content-Type": "application/text"})
        logger.info(f"Closed model '{self._model_id}'.")

    def get_modules(self, sort_by: SortBy = None, descending: bool = False) -> list[Module]:
        """
        Lists all the Modules in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Modules.
        """
        res = self._http.get_paginated(
            f"{self._url}/modules", "modules", params=sort_params(sort_by, descending)
        )
        return [Module.model_validate(e) for e in res]

    def get_views(
        self, sort_by: Literal["id", "module_id", "name"] | None = None, descending: bool = False
    ) -> list[View]:
        """
        Lists all the Views in the Model. This will include all Modules and potentially other saved
        views.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Views.
        """
        params = {"includesubsidiaryviews": True} | sort_params(sort_by, descending)
        return [
            View.model_validate(e)
            for e in self._http.get_paginated(f"{self._url}/views", "views", params=params)
        ]

    def get_view_info(self, view_id: int) -> ViewInfo:
        """
        Gets the detailed information about a View.
        :param view_id: The ID of the View.
        :return: The information about the View.
        """
        return ViewInfo.model_validate(self._http.get(f"{self._url}/views/{view_id}"))

    def get_line_items(self, only_module_id: int | None = None) -> list[LineItem]:
        """
        Lists all the Line Items in the Model.
        :param only_module_id: If provided, only Line Items from this Module will be returned.
        :return: All Line Items on this Model or only from the specified Module.
        """
        url = (
            f"{self._url}/modules/{only_module_id}/lineItems?includeAll=true"
            if only_module_id
            else f"{self._url}/lineItems?includeAll=true"
        )
        return [LineItem.model_validate(e) for e in self._http.get(url).get("items", [])]

    def get_lists(self, sort_by: SortBy = None, descending: bool = False) -> list[List]:
        """
        Lists all the Lists in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: All Lists on this model.
        """
        res = self._http.get_paginated(
            f"{self._url}/lists", "lists", params=sort_params(sort_by, descending)
        )
        return [List.model_validate(e) for e in res]

    def get_list_metadata(self, list_id: int) -> ListMetadata:
        """
        Gets the metadata for a List.
        :param list_id: The ID of the List.
        :return: The Metadata for the List.
        """
        return ListMetadata.model_validate(
            self._http.get(f"{self._url}/lists/{list_id}").get("metadata")
        )

    @overload
    def get_list_items(
        self, list_id: int, return_raw: Literal[False] = False
    ) -> list[ListItem]: ...

    @overload
    def get_list_items(
        self, list_id: int, return_raw: Literal[True] = True
    ) -> list[dict[str, Any]]: ...

    def get_list_items(
        self, list_id: int, return_raw: bool = False
    ) -> list[ListItem] | list[dict[str, Any]]:
        """
        Gets all the items in a List.
        :param list_id: The ID of the List.
        :param return_raw: If True, returns the items as a list of dictionaries instead of ListItem
               objects. If you use the result of this call in a DataFrame or you simply pass on the
               data, you will want to set this to avoid unnecessary (de-)serialization.
        :return: All items in the List.
        """
        res = self._http.get(f"{self._url}/lists/{list_id}/items?includeAll=true")
        if return_raw:
            return res.get("listItems", [])
        return [ListItem.model_validate(e) for e in res.get("listItems", [])]

    def insert_list_items(
        self, list_id: int, items: list[dict[str, str | int | dict]]
    ) -> InsertionResult:
        """
        Insert new items to the given list. The items must be a list of dictionaries with at least
        the keys `code` and `name`. You can optionally pass further keys for parents, extra
        properties etc. If you pass a long list, it will be split into chunks of 100,000 items, the
        maximum allowed by the API.

        **Warning**: If one or some of the requests timeout during large batch operations, the
        operation may actually complete on the server. Retries for these chunks will then report
        these items as "ignored" rather than "added", leading to misleading results. The results in
        Anaplan will be correct, but this function may report otherwise. Be generous with your
        timeouts and retries if you are using this function for large batch operations.

        :param list_id: The ID of the List.
        :param items: The items to insert into the List.
        :return: The result of the insertion, indicating how many items were added,
                 ignored or failed.
        """
        if not items:
            return InsertionResult(added=0, ignored=0, failures=[], total=0)
        if len(items) <= 100_000:
            result = InsertionResult.model_validate(
                self._http.post(
                    f"{self._url}/lists/{list_id}/items?action=add", json={"items": items}
                )
            )
            logger.info(f"Inserted {result.added} items into list '{list_id}'.")
            return result

        with ThreadPoolExecutor() as executor:
            responses = list(
                executor.map(
                    lambda chunk: self._http.post(
                        f"{self._url}/lists/{list_id}/items?action=add", json={"items": chunk}
                    ),
                    [items[i : i + 100_000] for i in range(0, len(items), 100_000)],
                )
            )
        result = parse_insertion_response(responses)
        logger.info(f"Inserted {result.added} items into list '{list_id}'.")
        return result

    def delete_list_items(
        self, list_id: int, items: list[dict[str, str | int]]
    ) -> ListDeletionResult:
        """
        Deletes items from a List. If you pass a long list, it will be split into chunks of 100,000
        items, the maximum allowed by the API.

        **Warning**: If one or some of the requests timeout during large batch operations, the
        operation may actually complete on the server. Retries for these chunks will then report
        none of these items as deleted, since on the retry none are removed, leading to misleading
        results. The results in Anaplan will be correct, but this function may report otherwise.
        Be generous with your timeouts and retries if you are using this function for large batch
        operations.

        :param list_id: The ID of the List.
        :param items: The items to delete from the List. Must be a dict with either `code` or `id`
                      as the keys to identify the records to delete. Specifying both will error.
        :return: The result of the deletion, indicating how many items were deleted or failed.
        """
        if not items:
            return ListDeletionResult(deleted=0, failures=[])
        if len(items) <= 100_000:
            res = self._http.post(
                f"{self._url}/lists/{list_id}/items?action=delete", json={"items": items}
            )
            info = ListDeletionResult.model_validate(res)
            logger.info(f"Deleted {info.deleted} items from list '{list_id}'.")
            return info

        with ThreadPoolExecutor() as executor:
            responses = list(
                executor.map(
                    lambda chunk: self._http.post(
                        f"{self._url}/lists/{list_id}/items?action=delete", json={"items": chunk}
                    ),
                    [items[i : i + 100_000] for i in range(0, len(items), 100_000)],
                )
            )
        info = ListDeletionResult(
            deleted=sum(res.get("deleted", 0) for res in responses),
            failures=list(chain.from_iterable(res.get("failures", []) for res in responses)),
        )
        logger.info(f"Deleted {info} items from list '{list_id}'.")
        return info

    def reset_list_index(self, list_id: int) -> None:
        """
        Resets the index of a List. The List must be empty to do so.
        :param list_id: The ID of the List.
        """
        self._http.post_empty(f"{self._url}/lists/{list_id}/resetIndex")
        logger.info(f"Reset index for list '{list_id}'.")

    def update_module_data(
        self, module_id: int, data: list[dict[str, Any]]
    ) -> int | dict[str, Any]:
        """
        Write the passed items to the specified module. If successful, the number of cells changed
        is returned, if only partially successful or unsuccessful, the response with the according
        details is returned instead.

        **You can update a maximum of 100,000 cells or 15 MB of data (whichever is lower) in a
        single request.** You must chunk your data accordingly. This is not done by this SDK,
        since it is discouraged. For larger imports, you should use the Bulk API instead.

        For more details see: https://anaplan.docs.apiary.io/#UpdateModuleCellData.
        :param module_id: The ID of the Module.
        :param data: The data to write to the Module.
        :return: The number of cells changed or the response with the according error details.
        """
        res = self._http.post(f"{self._url}/modules/{module_id}/data", json=data)
        if "failures" not in res:
            logger.info(f"Updated {res['numberOfCellsChanged']} cells in module '{module_id}'.")
        return res if "failures" in res else res["numberOfCellsChanged"]

    def get_current_period(self) -> CurrentPeriod:
        """
        Gets the current period of the model.
        :return: The current period of the model.
        """
        res = self._http.get(f"{self._url}/currentPeriod")
        return CurrentPeriod.model_validate(res["currentPeriod"])

    def set_current_period(self, date: str) -> CurrentPeriod:
        """
        Sets the current period of the model to the given date.
        :param date: The date to set the current period to, in the format 'YYYY-MM-DD'.
        :return: The updated current period of the model.
        """
        res = self._http.put(f"{self._url}/currentPeriod", {"date": date})
        logger.info(f"Set current period to '{date}'.")
        return CurrentPeriod.model_validate(res["currentPeriod"])

    def set_current_fiscal_year(self, year: str) -> FiscalYear:
        """
        Sets the current fiscal year of the model.
        :param year: The fiscal year to set, in the format specified in the model, e.g. FY24.
        :return: The updated fiscal year of the model.
        """
        res = self._http.put(f"{self._url}/modelCalendar/fiscalYear", {"year": year})
        logger.info(f"Set current fiscal year to '{year}'.")
        return FiscalYear.model_validate(res["modelCalendar"]["fiscalYear"])

    def get_model_calendar(self) -> ModelCalendar:
        """
        Get the calendar settings of the model.
        :return: The calendar settings of the model.
        """
        return parse_calendar_response(self._http.get(f"{self._url}/modelCalendar"))

    def get_dimension_items(self, dimension_id: int) -> list[DimensionWithCode]:
        """
        Get all items in a dimension. This will fail if the dimensions holds more than 1_000_000
        items. Valid Dimensions are:

        - Lists (101xxxxxxxxx)
        - List Subsets (109xxxxxxxxx)
        - Line Item Subsets (114xxxxxxxxx)
        - Users (101999999999)
        For lists and users, you should prefer using the `get_list_items` and `get_users` methods,
        respectively, instead.
        :param dimension_id: The ID of the dimension to list items for.
        :return: A list of Dimension items.
        """
        res = self._http.get(f"{self._url}/dimensions/{validate_dimension_id(dimension_id)}/items")
        return [DimensionWithCode.model_validate(e) for e in res.get("items", [])]

    def lookup_dimension_items(
        self, dimension_id: int, codes: list[str] = None, names: list[str] = None
    ) -> list[DimensionWithCode]:
        """
        Looks up items in a dimension by their codes or names. If both are provided, both will be
        searched for. You must provide at least one of `codes` or `names`. Valid Dimensions to
        lookup are:

        - Lists (101xxxxxxxxx)
        - Time (20000000003)
        - Version (20000000020)
        - Users (101999999999)
        :param dimension_id: The ID of the dimension to lookup items for.
        :param codes: A list of codes to lookup in the dimension.
        :param names: A list of names to lookup in the dimension.
        :return: A list of Dimension items that match the provided codes or names.
        """
        if not codes and not names:
            raise ValueError("At least one of 'codes' or 'names' must be provided.")
        if not (
            dimension_id == 101999999999
            or 101000000000 <= dimension_id < 102000000000
            or dimension_id == 20000000003
            or dimension_id == 20000000020
        ):
            raise InvalidIdentifierException(
                "Invalid dimension_id. Must be a List (101xxxxxxxxx), Time (20000000003), "
                "Version (20000000020), or Users (101999999999)."
            )
        res = self._http.post(
            f"{self._url}/dimensions/{dimension_id}/items", json={"codes": codes, "names": names}
        )
        return [DimensionWithCode.model_validate(e) for e in res.get("items", [])]

    def get_view_dimension_items(self, view_id: int, dimension_id: int) -> list[Dimension]:
        """
        Get the members of a dimension that are part of the given View. This call returns data as
        filtered by the page builder when they configure the view. This call respects hidden items,
        filtering selections, and Selective Access. If the view contains hidden or filtered items,
        these do not display in the response. This will fail if the dimensions holds more than
        1_000_000 items. The response returns Items within a flat list (no hierarchy) and order
        is not guaranteed.
        :param view_id: The ID of the View.
        :param dimension_id: The ID of the Dimension to get items for.
        :return: A list of Dimensions used in the View.
        """
        res = self._http.get(f"{self._url}/views/{view_id}/dimensions/{dimension_id}/items")
        return [Dimension.model_validate(e) for e in res.get("items", [])]

    def get_line_item_dimensions(self, line_item_id: int) -> list[Dimension]:
        """
        Get the dimensions of a Line Item. This will return all dimensions that are used in the
        Line Item.
        :param line_item_id: The ID of the Line Item.
        :return: A list of Dimensions used in the Line Item.
        """
        res = self._http.get(f"{self._url}/lineItems/{line_item_id}/dimensions")
        return [Dimension.model_validate(e) for e in res.get("dimensions", [])]



from typing import Literal, TypeAlias, Union

from pydantic import Field

from ._base import AnaplanModel


class User(AnaplanModel):
    id: str = Field(description="The unique identifier of this user.")
    active: bool = Field(description="Whether this user is active or not.")
    email: str = Field(description="The email address of this user.")
    email_opt_in: bool = Field(
        description="Whether this user has opted in to receive emails or not."
    )
    first_name: str = Field(description="The first name of this user.")
    last_name: str = Field(description="The last name of this user.")
    last_login_date: str | None = Field(
        None, description="The last login date of this user in ISO format."
    )


class ListItem(AnaplanModel):
    id: int = Field(description="The unique identifier of this list item.")
    name: str = Field(description="The name of this list item.")
    code: str | None = Field(None, description="The code of this list item.")
    properties: dict = Field({}, description="The properties of this list item.")
    subsets: dict = Field({}, description="The subsets of this list item.")
    parent: str | None = Field(None, description="The parent of this list item.")
    parent_id: str | None = Field(
        None, description="The unique identifier of the parent of this list item."
    )


class Module(AnaplanModel):
    id: int = Field(description="The unique identifier of this module.")
    name: str = Field(description="The name of this module.")


class Dimension(AnaplanModel):
    id: int = Field(description="The unique identifier of this dimension.")
    name: str = Field(description="The name of this dimension.")


class DimensionWithCode(Dimension):
    code: str = Field(description="The code of this dimension.")


class View(AnaplanModel):
    code: str = Field(description="The code of this view.")
    id: int = Field(description="The unique identifier of this view.")
    name: str = Field(description="The name of this views.")
    moduleId: int = Field(description="The unique identifier of the module this view belongs to.")


class ViewInfo(AnaplanModel):
    view_id: int = Field(description="The unique identifier of this view.")
    view_name: str = Field(description="The name of this view.")
    rows: list[Dimension] = Field(
        [], description="The list of dimensions in the rows of this view."
    )
    pages: list[Dimension] = Field(
        [], description="The list of dimensions in the pages of this view."
    )


class PeriodType(AnaplanModel):
    entity_id: Literal["YEAR", "HALF_YEAR", "MONTH", "QUARTER", "WEEK", "DAY"] = Field(
        description="The type of period entity."
    )
    entity_label: Literal["Year", "Half-Year", "Month", "Quarter", "Week", "Day"] = Field(
        description="The type of period entity."
    )
    entity_index: int = Field(description="The index of the period entity")


class EntityFormatFilter(AnaplanModel):
    source_line_item_or_property: str = Field(
        description="The unique identifier of the source line item or property."
    )
    mapping_hierarchy: str = Field(description="The unique identifier of the mapping hierarchy.")
    key_property: str = Field(description="The unique identifier of the key property.")
    value_property: str = Field(description="The unique identifier of the value property.")


class TextMetadata(AnaplanModel):
    data_type: Literal["TEXT"] = Field(
        description="The data type. Literal for the tagged union discriminator."
    )
    text_type: Literal["DRILLTHRU_URI", "EMAIL_ADDRESS", "GENERAL"] = Field(
        description="The text type."
    )


class ListMetadata(AnaplanModel):
    data_type: Literal["ENTITY"] = Field(
        description="The data type. Literal for the tagged union discriminator."
    )
    hierarchy_entity_id: int = Field(
        validation_alias="hierarchyEntityLongId",
        description="The unique identifier of the hierarchy entity, like Lists or List Subsets.",
    )
    selective_access_applied: bool = Field(
        description="Whether selective access is applied or not."
    )
    show_all: bool = Field(description="Whether to show all values or not.")
    entity_format_filter: EntityFormatFilter | None = Field(
        None, description="Entity format filter configuration."
    )


class TimePeriodMetadata(AnaplanModel):
    data_type: Literal["TIME_ENTITY"] = Field(
        description="The data type. Literal for the tagged union discriminator."
    )
    period_type: PeriodType = Field(description="The period type.")


class NumberMetadata(AnaplanModel):
    data_type: Literal["NUMBER"] = Field(
        description="The data type. Literal for the tagged union discriminator."
    )
    comparison_increase: Literal["GOOD", "BAD", "NEUTRAL"] | None = Field(
        description="The comparison increase setting."
    )
    custom_units: str | None = Field(None, description="Custom units for display.")
    decimal_places: int = Field(description="Number of decimal places.")
    decimal_separator: Literal["COMMA", "FULL_STOP"] = Field(description="The decimal separator.")
    units_display_type: Literal[
        "CUSTOM_SUFFIX",
        "CUSTOM_PREFIX",
        "CURRENCY_CODE",
        "CURRENCY_SYMBOL",
        "PERCENTAGE_SUFFIX",
        "NONE",
    ] = Field(description="Units display type.")
    units_type: Literal["CUSTOM", "CURRENCY", "PERCENTAGE", "NONE"] = Field(
        description="Units type."
    )
    zero_format: Literal["HYPHEN", "ZERO", "BLANK"] = Field(description="Zero format display.")
    grouping_separator: Literal["COMMA", "FULL_STOP", "SPACE", "NONE"] = Field(
        description="The grouping separator."
    )
    minimum_significant_digits: int = Field(description="Minimum significant digits.")
    negative_number_notation: Literal["MINUS_SIGN", "PARENTHESES"] = Field(
        description="Negative number notation."
    )


class GenericTypeMetadata(AnaplanModel):
    data_type: Literal["BOOLEAN", "NONE", "DATE"] = Field(description="The data type.")


class LineItem(AnaplanModel):
    id: int = Field(description="The unique identifier of this line item.")
    name: str = Field(description="The name of this line item.")
    module_id: int = Field(
        description="The unique identifier of the module this line item belongs to."
    )
    module_name: str = Field(description="The name of the module this line item belongs to.")
    format: Literal["NUMBER", "BOOLEAN", "TEXT", "NONE", "DATE", "LIST", "TIME PERIOD"] = Field(
        description="The format of this line item."
    )
    format_metadata: (
        NumberMetadata | ListMetadata | TimePeriodMetadata | TextMetadata | GenericTypeMetadata
    ) = Field(
        description="The format metadata of this line item. Each Type provides different metadata.",
        discriminator="data_type",
    )
    summary: str = Field(description="The summary of this line item.")
    applies_to: list[Dimension] = Field([], description="The applies to value of this line item.")
    data_tags: list[Dimension] = Field([], description="The data tags of this line item.")
    referenced_by: list[Dimension] = Field([], description="List of references to this line item.")
    time_scale: str = Field(description="The time scale of this line item.")
    time_range: str = Field(description="The time range of this line item.")
    version: Dimension = Field(description="The version of this line item.")
    parent: Dimension | None = Field(None, description="The Parent of this line item.")
    read_access_driver: Dimension | None = Field(
        None, description="The read access driver of this line item."
    )
    write_access_driver: Dimension | None = Field(
        None, description="The write access driver of this line item."
    )
    style: str = Field(description="The style of this line item.")
    cell_count: int | None = Field(None, description="The cell count of this line item.")
    notes: str = Field(description="The notes of this line item.")
    code: str | None = Field(None, description="The code of this line item.")
    is_summary: bool = Field(description="Whether this line item is a summary or not.")
    formula: str | None = Field(None, description="The formula of this line item.")
    formula_scope: str = Field(description="The formula scope of this line item.")
    use_switchover: bool = Field(description="Whether the switchover is used or not.")
    breakback: bool = Field(description="Whether the breakback is enabled or not.")
    brought_forward: bool = Field(description="Whether the brought forward is enabled or not.")
    start_of_section: bool = Field(
        description="Whether this line item is the start of a section or not."
    )


class Failure(AnaplanModel):
    index: int = Field(
        validation_alias="requestIndex", description="The index of the item that failed."
    )
    reason: str = Field(validation_alias="failureType", description="The reason for the failure.")
    details: str = Field(
        validation_alias="failureMessageDetails", description="The details of the failure."
    )


class ModelStatus(AnaplanModel):
    peak_memory_usage_estimate: int | None = Field(
        description="The peak memory usage estimate of this model."
    )
    peak_memory_usage_time: int | None = Field(
        description="The peak memory usage time of this model."
    )
    progress: float = Field(description="The progress of this model.")
    current_step: str = Field(description="The current step of this model.")
    tooltip: str | None = Field(description="The tooltip of this model.")
    task_id: str | None = Field(description="The unique identifier of the task of this model.")
    creation_time: int = Field(description="The creation time of this model.")
    export_task_type: str | None = Field(description="The export task type of this model.")


class InsertionResult(AnaplanModel):
    added: int = Field(description="The number of items successfully added.")
    ignored: int = Field(description="The number of items ignored, or items that failed.")
    total: int = Field(description="The total number of items.")
    failures: list[Failure] = Field([], description="The list of failures.")


class ListDeletionResult(AnaplanModel):
    deleted: int = Field(description="The number of items successfully deleted.")
    failures: list[Failure] = Field([], description="The list of failures.")


class PartialCurrentPeriod(AnaplanModel):
    period_text: str = Field(description="The text representation of the current period.")
    last_day: str = Field(description="The last day of the current period in YYYY-MM-DD format.")


class CurrentPeriod(PartialCurrentPeriod):
    calendar_type: str = Field(description="The type of calendar used for the current period.")


class FiscalYear(AnaplanModel):
    year: str = Field(description="The fiscal year in the format set in the model, e.g. FY24.")
    start_date: str = Field(description="The start date of the fiscal year in YYYY-MM-DD format.")
    end_date: str = Field(description="The end date of the fiscal year in YYYY-MM-DD format.")


class TotalsSelection(AnaplanModel):
    quarter_totals: bool = Field(description="Whether quarter totals are enabled.")
    half_year_totals: bool = Field(description="Whether half year totals are enabled.")
    year_to_date_summary: bool = Field(description="Whether year to date summary is enabled.")
    year_to_go_summary: bool = Field(description="Whether year to go summary is enabled.")
    total_of_all_periods: bool = Field(description="Whether total of all periods is enabled.")


class TotalsSelectionWithQuarter(TotalsSelection):
    extra_month_quarter: Literal[1, 2, 3, 4] = Field(
        description="The quarter in which the extra month is included."
    )


class BaseCalendar(AnaplanModel):
    calendar_type: Literal[
        "Calendar Months/Quarters/Years",
        "Weeks: 4-4-5, 4-5-4 or 5-4-4",
        "Weeks: General",
        "Weeks: 13 4-week Periods",
    ] = Field(description="The type of calendar used.")
    current_period: PartialCurrentPeriod = Field(description="The current period configuration.")


class MonthsQuartersYearsCalendar(BaseCalendar):
    past_years_count: int = Field(description="The number of past years included.")
    fiscal_year: FiscalYear = Field(description="The fiscal year configuration.")
    totals_selection: TotalsSelection = Field(
        description="The totals selection configuration for the calendar."
    )


class WeeksGeneralCalendar(BaseCalendar):
    start_date: str = Field(description="The start date of the calendar in YYYY-MM-DD format.")
    weeks_count: int = Field(description="The number of weeks in the calendar.")


class WeeksPeriodsCalendar(BaseCalendar):
    fiscal_year: FiscalYear = Field(description="The fiscal year configuration.")
    past_years_count: int = Field(description="The number of past years included.")
    future_years_count: int = Field(description="The number of future years included.")
    extra_week_month: int = Field(description="The month in which the extra week is included.")
    week_format: Literal["Numbered", "Week Commencing", "Week Ending"] = Field(
        description="The format of the week."
    )
    totals_selection: TotalsSelectionWithQuarter = Field(
        description="The totals selection configuration for the calendar."
    )


class WeeksGroupingCalendar(WeeksPeriodsCalendar):
    week_grouping: str = Field(
        description="The week grouping configuration, e.g. '4-4-5', '4-5-4', or '5-4-4'."
    )
    totals_selection: TotalsSelection = Field(
        description="The totals selection configuration for the calendar."
    )


ModelCalendar: TypeAlias = Union[
    MonthsQuartersYearsCalendar, WeeksGeneralCalendar, WeeksGroupingCalendar, WeeksPeriodsCalendar
]
